@@ -95,20 +95,14 @@ struct logmsg {
 	char str[0];
 };
 
+typedef int (*formatter_fn)(char *, size_t, const struct logmsg *, bool);
+
 struct log_format {
 	const char *name;
-	int (*formatter)(char *, size_t, const struct logmsg *);
-	struct list_head list;
+	formatter_fn formatter;
+	struct list_node list;
 };
 
-#define log_format_register(n, formatter_fn)				\
-	static void __attribute__((constructor(101)))			\
-	regist_ ## formatter_fn(void) {					\
-		static struct log_format f =				\
-			{ .name = n, .formatter = formatter_fn };	\
-		list_add(&f.list, &log_formats);			\
-}
-
 static LIST_HEAD(log_formats);
 static struct log_format *format;
 
@@ -120,13 +114,44 @@ static const char *log_name;
 static char *log_nowname;
 int sd_log_level = SDOG_INFO;
 static pid_t sheep_pid;
-static pid_t logger_pid;
+pid_t logger_pid = -1;
 static key_t semkey;
 static char *log_buff;
 
 static int64_t max_logsize = 500 * 1024 * 1024;  /*500MB*/
 
-static pthread_mutex_t logsize_lock = PTHREAD_MUTEX_INITIALIZER;
+static enum log_dst_type dst_type = LOG_DST_STDOUT;
+
+/*
+ * block_sighup()
+ *
+ * used for protecting log_fd from SIGHUP rotation
+ */
+static void block_sighup(void)
+{
+	int ret;
+	sigset_t new, old;
+
+	sigemptyset(&new);
+	sigemptyset(&old);
+	sigaddset(&new, SIGHUP);
+	ret = sigprocmask(SIG_BLOCK, &new, &old);
+	if (ret < 0)
+		syslog(LOG_ERR, "blocking SIGHUP failed\n");
+}
+
+static void unblock_sighup(void)
+{
+	int ret;
+	sigset_t new, old;
+
+	sigemptyset(&new);
+	sigemptyset(&old);
+	sigaddset(&new, SIGHUP);
+	ret = sigprocmask(SIG_UNBLOCK, &new, &old);
+	if (ret < 0)
+		syslog(LOG_ERR, "unblock SIGHUP failed\n");
+}
 
 static const char *format_thread_name(char *str, size_t size, const char *name,
 				      int idx)
@@ -141,15 +166,116 @@ static const char *format_thread_name(char *str, size_t size, const char *name,
 	return str;
 }
 
+static int server_log_formatter(char *buff, size_t size,
+				const struct logmsg *msg, bool print_time)
+{
+	char *p = buff;
+	struct tm tm;
+	size_t len;
+	char thread_name[MAX_THREAD_NAME_LEN];
+
+	if (print_time) {
+		localtime_r(&msg->tv.tv_sec, &tm);
+		len = strftime(p, size, "%b %2d %H:%M:%S ",
+			       (const struct tm *)&tm);
+		p += len;
+		size -= len;
+	}
+
+	len = snprintf(p, size, "%s%6s %s[%s] %s(%d) %s%s%s",
+		       colorize ? log_color[msg->prio] : "",
+		       log_prio_str[msg->prio],
+		       colorize ? TEXT_YELLOW : "",
+		       format_thread_name(thread_name, sizeof(thread_name),
+					  msg->worker_name, msg->worker_idx),
+		       msg->func, msg->line,
+		       colorize ? log_color[msg->prio] : "",
+		       msg->str, colorize ? TEXT_NORMAL : "");
+	if (len < 0)
+		len = 0;
+	p += min(len, size - 1);
+
+	return p - buff;
+}
+
+static int default_log_formatter(char *buff, size_t size,
+				 const struct logmsg *msg, bool print_time)
+{
+	size_t len = min(size, msg->str_len);
+
+	memcpy(buff, msg->str, len);
+
+	return len;
+}
+
+static int json_log_formatter(char *buff, size_t size,
+			      const struct logmsg *msg, bool print_time)
+{
+	char *p = buff;
+	size_t len;
+
+	assert(logger_user_info);
+
+	len = snprintf(p, size, "{ \"user_info\": "
+		       "{\"program_name\": \"%s\", \"port\": %d},"
+		       "\"body\": {"
+		       "\"second\": %lu, \"usecond\": %lu, "
+		       "\"worker_name\": \"%s\", \"worker_idx\": %d, "
+		       "\"func\": \"%s\", \"line\": %d, "
+		       "\"msg\": \"",
+		       log_name, logger_user_info->port,
+		       msg->tv.tv_sec, msg->tv.tv_usec,
+		       msg->worker_name[0] ? msg->worker_name : "main",
+		       msg->worker_idx, msg->func, msg->line);
+	if (len < 0)
+		return 0;
+	len = min(len, size - 1);
+	p += len;
+	size -= len;
+
+	for (int i = 0; i < msg->str_len; i++) {
+		if (size <= 1)
+			break;
+
+		if (msg->str[i] == '"') {
+			*p++ = '\\';
+			size--;
+		}
+
+		if (size <= 1)
+			break;
+		*p++ = msg->str[i];
+		size--;
+	}
+
+	pstrcpy(p, size, "\"} }");
+	p += strlen(p);
+
+	return p - buff;
+}
+
+static void log_format_register(const char *name, formatter_fn formatter)
+{
+	struct log_format *f = xmalloc(sizeof(struct log_format));
+
+	f->name = name;
+	f->formatter = formatter;
+	list_add(&f->list, &log_formats);
+}
+
 /*
  * We need to set default log formatter because dog doesn't want to call
  * select_log_formatter().
  */
-static void __attribute__((constructor(65535)))
+static void __attribute__((constructor))
 init_log_formatter(void)
 {
 	struct log_format *f;
 
+	log_format_register("json", json_log_formatter);
+	log_format_register("server", server_log_formatter);
+	log_format_register("default", default_log_formatter);
+
 	list_for_each_entry(f, &log_formats, list) {
 		if (!strcmp(f->name, "default")) {
 			format = f;
@@ -230,106 +356,26 @@ static void free_logarea(void)
 	shmdt(la);
 }
 
-static int server_log_formatter(char *buff, size_t size,
-				const struct logmsg *msg)
-{
-	char *p = buff;
-	struct tm tm;
-	size_t len;
-	char thread_name[MAX_THREAD_NAME_LEN];
-
-	localtime_r(&msg->tv.tv_sec, &tm);
-	len = strftime(p, size, "%b %2d %H:%M:%S ", (const struct tm *)&tm);
-	p += len;
-	size -= len;
-
-	len = snprintf(p, size, "%s%6s %s[%s] %s(%d) %s%s%s",
-		       colorize ? log_color[msg->prio] : "",
-		       log_prio_str[msg->prio],
-		       colorize ? TEXT_YELLOW : "",
-		       format_thread_name(thread_name, sizeof(thread_name),
-					  msg->worker_name, msg->worker_idx),
-		       msg->func, msg->line,
-		       colorize ? log_color[msg->prio] : "",
-		       msg->str, colorize ? TEXT_NORMAL : "");
-	if (len < 0)
-		len = 0;
-	p += min(len, size - 1);
-
-	return p - buff;
-}
-log_format_register("server", server_log_formatter);
-
-static int default_log_formatter(char *buff, size_t size,
-				 const struct logmsg *msg)
-{
-	size_t len = min(size, msg->str_len);
-
-	memcpy(buff, msg->str, len);
-
-	return len;
-}
-log_format_register("default", default_log_formatter);
-
-static int json_log_formatter(char *buff, size_t size,
-				const struct logmsg *msg)
-{
-	char *p = buff;
-	size_t len;
-
-	assert(logger_user_info);
-
-	len = snprintf(p, size, "{ \"user_info\": "
-		       "{\"program_name\": \"%s\", \"port\": %d},"
-		       "\"body\": {"
-		       "\"second\": %lu, \"usecond\": %lu, "
-		       "\"worker_name\": \"%s\", \"worker_idx\": %d, "
-		       "\"func\": \"%s\", \"line\": %d, "
-		       "\"msg\": \"",
-		       log_name, logger_user_info->port,
-		       msg->tv.tv_sec, msg->tv.tv_usec,
-		       msg->worker_name[0] ? msg->worker_name : "main",
-		       msg->worker_idx, msg->func, msg->line);
-	if (len < 0)
-		return 0;
-	len = min(len, size - 1);
-	p += len;
-	size -= len;
-
-	for (int i = 0; i < msg->str_len; i++) {
-		if (size <= 1)
-			break;
-
-		if (msg->str[i] == '"') {
-			*p++ = '\\';
-			size--;
-		}
-
-		if (size <= 1)
-			break;
-		*p++ = msg->str[i];
-		size--;
-	}
-
-	pstrcpy(p, size, "\"} }");
-	p += strlen(p);
-
-	return p - buff;
-}
-log_format_register("json", json_log_formatter);
-
 /* this one can block under memory pressure */
 static void log_syslog(const struct logmsg *msg)
 {
 	char str[MAX_MSG_SIZE];
 	int len;
 
-	len = format->formatter(str, sizeof(str) - 1, msg);
-	str[len++] = '\n';
+	len = format->formatter(str, sizeof(str) - 1, msg, log_fd >= 0);
+	if (dst_type == LOG_DST_DEFAULT)
+		str[len++] = '\n';
+	else	/* LOG_DST_SYSLOG */
+		str[len++] = '\0';
+
+	block_sighup();
+
 	if (log_fd >= 0)
 		xwrite(log_fd, str, len);
 	else
 		syslog(msg->prio, "%s", str);
+
+	unblock_sighup();
 }
 
 static void init_logmsg(struct logmsg *msg, struct timeval *tv,
@@ -396,7 +442,8 @@ static void dolog(int prio, const char *func, int line,
 		char str_final[MAX_MSG_SIZE];
 
 		init_logmsg(msg, &tv, prio, func, line);
-		len = format->formatter(str_final, sizeof(str_final) - 1, msg);
+		len = format->formatter(str_final, sizeof(str_final) - 1, msg,
+					true);
 		str_final[len++] = '\n';
 		xwrite(fileno(stderr), str_final, len);
 		fflush(stderr);
@@ -505,16 +552,23 @@ static void crash_handler(int signo)
 	reraise_crash_signal(signo, 1);
 }
 
+static void sighup_handler(int signo)
+{
+	rotate_log();
+}
+
 static void logger(char *log_dir, char *outfile)
 {
 	int fd;
 
 	log_buff = xzalloc(la->end - la->start);
 
-	log_fd = open(outfile, O_CREAT | O_RDWR | O_APPEND, 0644);
-	if (log_fd < 0) {
-		syslog(LOG_ERR, "failed to open %s\n", outfile);
-		exit(1);
+	if (dst_type == LOG_DST_DEFAULT) {
+		log_fd = open(outfile, O_CREAT | O_RDWR | O_APPEND, 0644);
+		if (log_fd < 0) {
+			syslog(LOG_ERR, "failed to open %s\n", outfile);
+			exit(1);
+		}
 	}
 	la->active = true;
 
@@ -535,9 +589,7 @@ static void logger(char *log_dir, char *outfile)
 
 	/* flush when either the logger or its parent dies */
 	install_crash_handler(crash_handler);
-	install_sighandler(SIGHUP, crash_handler, false);
-
-	prctl(PR_SET_PDEATHSIG, SIGHUP);
+	install_sighandler(SIGHUP, sighup_handler, false);
 
 	/*
 	 * we need to check the aliveness of the sheep process since
@@ -549,10 +601,11 @@ static void logger(char *log_dir, char *outfile)
 	while (la->active) {
 		log_flush();
 
-		if (max_logsize) {
+		block_sighup();
+
+		if (dst_type == LOG_DST_DEFAULT && max_logsize) {
 			off_t offset;
 
-			pthread_mutex_lock(&logsize_lock);
 			offset = lseek(log_fd, 0, SEEK_END);
 			if (offset < 0) {
 				syslog(LOG_ERR, "sheep log error\n");
@@ -561,9 +614,14 @@ static void logger(char *log_dir, char *outfile)
 				if (log_size >= max_logsize)
 					rotate_log();
 			}
-			pthread_mutex_unlock(&logsize_lock);
 		}
 
+		unblock_sighup();
+
+		if (getppid() == 1)
+			/* My parent (sheep process) is dead. */
+			break;
+
 		sleep(1);
 	}
 
@@ -595,12 +653,13 @@ void early_log_init(const char *format_name, struct logger_user_info *user_info)
 	exit(1);
 }
 
-int log_init(const char *program_name, bool to_stdout, int level,
+int log_init(const char *program_name, enum log_dst_type type, int level,
 		     char *outfile)
 {
 	char log_dir[PATH_MAX], tmp[PATH_MAX];
 	int size = level == SDOG_DEBUG ? LOG_SPACE_DEBUG_SIZE : LOG_SPACE_SIZE;
 
+	dst_type = type;
 	sd_log_level = level;
 
 	log_name = program_name;
@@ -610,10 +669,15 @@ int log_init(const char *program_name, bool to_stdout, int level,
 
 	semkey = random();
 
-	if (to_stdout) {
+	switch (type) {
+	case LOG_DST_STDOUT:
 		if (is_stdout_console())
 			colorize = true;
-	} else {
+		break;
+	case LOG_DST_SYSLOG:
+		openlog(program_name, LOG_PID, LOG_DAEMON);
+		/* fall through */
+	case LOG_DST_DEFAULT:
 		if (logarea_init(size)) {
 			syslog(LOG_ERR, "failed to initialize the logger\n");
 			return 1;
@@ -636,6 +700,10 @@ int log_init(const char *program_name, bool to_stdout, int level,
 			syslog(LOG_WARNING, "logger pid %d starting\n", logger_pid);
 		else
 			logger(log_dir, outfile);
+		break;
+	default:
+		sd_err("unknown type of log destination type: %d", type);
+		return -1;
 	}
 
 	return 0;
@@ -668,145 +736,56 @@ void get_thread_name(char *name)
 
 #define SD_MAX_STACK_DEPTH 1024
 
-static int get_my_path(char *path, size_t size)
-{
-	/* readlink doesn't append '\0', so initialize here */
-	memset(path, 0, size);
-
-	return readlink("/proc/self/exe", path, size);
-}
-
 static bool check_gdb(void)
 {
 	return system("which gdb > /dev/null") == 0;
 }
 
-/*
- * __builtin_frame_address() returns address in frame pointer register if any
- * (e.g, in x86 it returns EBP). If no dedicated register, the frame address is
- * normally the address of the first word pushed on to the stack by the function
- *
- * For a normal subroutine setup, above the value __builtin_frame_address
- * returns, there are two addresses, which stores old EBP and old EIP, being
- * pushed on to the stack. So we have to plus 2 to get the right value for the
- * frame address, which is expected by GDB.
- *
- * This is tested on X86, other architetures aren't tested. But even if this
- * formula is wrong, GDB just doesn't procude anything useful after panic.
- */
-#define FRAME_POINTER ((unsigned long *)__builtin_frame_address(0) + 2)
+#define SD_ARG_MAX (sysconf(_SC_ARG_MAX))
 
-__attribute__ ((__noinline__))
-int __sd_dump_variable(const char *var)
+static int gdb_cmd(const char *cmd)
 {
-	char cmd[ARG_MAX], path[PATH_MAX], info[256];
-	FILE *f = NULL;
-	void *base_sp = FRAME_POINTER;
+	char time_str[256], cmd_str[SD_ARG_MAX];
+	time_t ti;
+	struct tm tm;
 
 	if (!check_gdb()) {
 		sd_debug("cannot find gdb");
 		return -1;
 	}
 
-	if (get_my_path(path, sizeof(path)) < 0)
-		return -1;
+	time(&ti);
+	localtime_r(&ti, &tm);
+	strftime(time_str, sizeof(time_str), "%b %2d %H:%M:%S ", &tm);
 
-	snprintf(cmd, sizeof(cmd), "gdb -nw %s %d -batch -ex 'set width 80'"
-		 " -ex 'select-frame %p' -ex 'up 1' -ex 'p %s' 2> /dev/null",
-		 path, gettid(), base_sp, var);
-	f = popen(cmd, "r");
-	if (f == NULL) {
-		sd_err("failed to run gdb");
-		return -1;
-	}
+	snprintf(cmd_str, sizeof(cmd_str),
+		 "gdb -nw %s %d -batch >/dev/null 2>&1"
+		 " -ex 'set logging on'"
+		 " -ex 'echo \\n'"
+		 " -ex 'echo ==\\n'"
+		 " -ex 'echo == %s\\n'"
+		 " -ex 'echo == program: %s\\n'"
+		 " -ex 'echo == command: %s\\n'"
+		 " -ex 'echo ==\\n'"
+		 " -ex '%s'"
+		 " -ex 'set logging off'",
+		 my_exe_path(), getpid(), time_str, my_exe_path(), cmd, cmd);
 
-	/*
-	 * The expected outputs of gdb are:
-	 *
-	 *  [some info we don't need]
-	 *  $1 = {
-	 *    <variable info>
-	 *  }
-	 */
-	sd_emerg("dump %s", var);
-	while (fgets(info, sizeof(info), f) != NULL) {
-		if (info[0] == '$') {
-			sd_emerg("%s", info);
-			break;
-		}
-	}
-	while (fgets(info, sizeof(info), f) != NULL)
-		sd_emerg("%s", info);
+	return system(cmd_str);
+}
 
-	pclose(f);
-	return 0;
+int __sd_dump_variable(const char *var)
+{
+	char cmd[256];
+
+	snprintf(cmd, sizeof(cmd), "p %s", var);
+
+	return gdb_cmd(cmd);
 }
 
-__attribute__ ((__noinline__))
 static int dump_stack_frames(void)
 {
-	char path[PATH_MAX];
-	int i, stack_no = 0;
-	void *base_sp = FRAME_POINTER;
-
-	if (!check_gdb()) {
-		sd_debug("cannot find gdb");
-		return -1;
-	}
-
-	if (get_my_path(path, sizeof(path)) < 0)
-		return -1;
-
-	for (i = 1; i < SD_MAX_STACK_DEPTH; i++) {
-		char cmd[ARG_MAX], info[256];
-		FILE *f = NULL;
-		bool found = false;
-
-		snprintf(cmd, sizeof(cmd), "gdb -nw %s %d -batch"
-			 " -ex 'set width 80' -ex 'select-frame %p'"
-			 " -ex 'up %d' -ex 'info locals' 2> /dev/null",
-			 path, gettid(), base_sp, i);
-		f = popen(cmd, "r");
-		if (f == NULL)
-			return -1;
-		/*
-		 * The expected outputs of gdb are:
-		 *
-		 *  [some info we don't need]
-		 *  #<stack no> <addr> in <func>(<arg>) at <file>:<line>
-		 *  <line>   <source>
-		 *  <local variables>
-		 */
-		while (fgets(info, sizeof(info), f) != NULL) {
-			int no;
-			if (sscanf(info, "#%d ", &no) == 1) {
-				if (no <= stack_no) {
-					/* reached to the end of the stacks */
-					pclose(f);
-					return 0;
-				}
-				stack_no = no;
-				found = true;
-				sd_emerg("%s", info);
-				break;
-			}
-		}
-
-		if (!found) {
-			sd_info("Cannot get info from GDB");
-			sd_info("Set /proc/sys/kernel/yama/ptrace_scope to"
-				" zero if you are using Ubuntu.");
-			pclose(f);
-			return -1;
-		}
-
-		while (fgets(info, sizeof(info), f) != NULL)
-			sd_emerg("%s", info);
-
-		pclose(f);
-	}
-
-	return 0;
+	return gdb_cmd("thread apply all where full");
 }
 
 __attribute__ ((__noinline__))
@@ -817,7 +796,7 @@ void sd_backtrace(void)
 
 	for (i = 1; i < n; i++) { /* addrs[0] is here, so skip it */
 		void *addr = addrs[i];
-		char cmd[ARG_MAX], path[PATH_MAX], info[256], **str;
+		char cmd[SD_ARG_MAX], info[256], **str;
 		FILE *f;
 
 		/*
@@ -827,12 +806,9 @@ void sd_backtrace(void)
 		addr = (void *)((char *)addr - 1);
 
 		/* try to get a line number with addr2line if possible */
-		if (get_my_path(path, sizeof(path)) < 0)
-			goto fallback;
-
 		snprintf(cmd, sizeof(cmd), "addr2line -s -e %s -f -i %p | "
-			"perl -e '@a=<>; chomp @a; print \"$a[1]: $a[0]\"'",
-			path, addr);
+			 "perl -e '@a=<>; chomp @a; print \"$a[1]: $a[0]\"'",
+			 my_exe_path(), addr);
 		f = popen(cmd, "r");
 		if (!f)
 			goto fallback;
@@ -840,7 +816,7 @@ void sd_backtrace(void)
 			goto fallback_close;
 
 		if (info[0] != '?' && info[0] != '\0')
-			sd_emerg("%s", info);
+			sd_emerg("%s", chomp(info));
 		else
 			goto fallback_close;
 
@@ -861,3 +837,14 @@ fallback:
 	/* dump the stack frames if possible*/
 	dump_stack_frames();
 }
+
+void set_loglevel(int new_loglevel)
+{
+	assert(SDOG_EMERG <= new_loglevel && new_loglevel <= SDOG_DEBUG);
+	sd_log_level = new_loglevel;
+}
+
+int get_loglevel(void)
+{
+	return sd_log_level;
+}
