@@ -14,7 +14,6 @@
 #include <string.h>
 #include <unistd.h>
 #include <sys/epoll.h>
-#include <sys/eventfd.h>
 #include <zookeeper/zookeeper.h>
 #include <pthread.h>
 
@@ -31,6 +30,39 @@
 #define QUEUE_ZNODE BASE_ZNODE "/queue"
 #define MEMBER_ZNODE BASE_ZNODE "/member"
 #define MASTER_ZNONE BASE_ZNODE "/master"
+#define LOCK_ZNODE BASE_ZNODE "/lock"
+
+static int zk_timeout = SESSION_TIMEOUT;
+
+/* structure for distributed lock */
+struct cluster_lock {
+	struct hlist_node hnode;
+	/* id is passed by users to represent a lock handle */
+	uint64_t id;
+	/* referenced by different threads in one sheepdog daemon */
+	uint64_t ref;
+	/* wait for the release of id by other lock owner */
+	struct sd_cond wait_wakeup;
+	struct sd_mutex wait_mutex;
+	/* lock for different threads of the same node on the same id */
+	struct sd_mutex id_lock;
+	char lock_path[MAX_NODE_STR_LEN];
+};
+
+#define WAIT_TIME	1		/* second */
+
+#define HASH_BUCKET_NR	1021
+static struct hlist_head *cluster_locks_table;
+static struct sd_mutex table_locks[HASH_BUCKET_NR];
+
+/*
+ * Wait a while when create, delete or get_children fail on
+ * zookeeper lock so it will not print too much loop log
+ */
+static void zk_wait(void)
+{
+	sleep(WAIT_TIME);
+}
 
 /* iterate child znodes */
 #define FOR_EACH_ZNODE(parent, path, strs)			       \
@@ -51,13 +83,15 @@ enum zk_event_type {
 };
 
 struct zk_node {
-	struct list_head list;
+	struct list_node list;
 	struct rb_node rb;
 	struct sd_node node;
 	bool callbacked;
 	bool gone;
 };
 
+#define ZK_MAX_BUF_SIZE (1*1024*1024) /* 1M */
+
 struct zk_event {
 	uint64_t id;
 	enum zk_event_type type;
@@ -65,14 +99,14 @@ struct zk_event {
 	size_t msg_len;
 	size_t nr_nodes;
 	size_t buf_len;
-	uint8_t buf[SD_MAX_EVENT_BUF_SIZE];
+	uint8_t buf[ZK_MAX_BUF_SIZE];
 };
 
-static struct sd_node sd_nodes[SD_MAX_NODES];
+static struct rb_root sd_node_root = RB_ROOT;
 static size_t nr_sd_nodes;
 static struct rb_root zk_node_root = RB_ROOT;
-static struct sd_lock zk_tree_lock = SD_LOCK_INITIALIZER;
-static struct sd_lock zk_compete_master_lock = SD_LOCK_INITIALIZER;
+static struct sd_rw_lock zk_tree_lock = SD_RW_LOCK_INITIALIZER;
+static struct sd_rw_lock zk_compete_master_lock = SD_RW_LOCK_INITIALIZER;
 static LIST_HEAD(zk_block_list);
 static uatomic_bool is_master;
 static uatomic_bool stop;
@@ -81,51 +115,21 @@ static bool first_push = true;
 
 static void zk_compete_master(void);
 
+static int zk_node_cmp(const struct zk_node *a, const struct zk_node *b)
+{
+	return node_id_cmp(&a->node.nid, &b->node.nid);
+}
+
 static struct zk_node *zk_tree_insert(struct zk_node *new)
 {
-	struct rb_node **p = &zk_node_root.rb_node;
-	struct rb_node *parent = NULL;
-	struct zk_node *entry;
-
-	while (*p) {
-		int cmp;
-
-		parent = *p;
-		entry = rb_entry(parent, struct zk_node, rb);
-
-		cmp = node_cmp(&new->node, &entry->node);
-		if (cmp < 0)
-			p = &(*p)->rb_left;
-		else if (cmp > 0)
-			p = &(*p)->rb_right;
-		else
-			/* already has this entry */
-			return entry;
-	}
-	rb_link_node(&new->rb, parent, p);
-	rb_insert_color(&new->rb, &zk_node_root);
-	return NULL; /* insert successfully */
+	return rb_insert(&zk_node_root, new, rb, zk_node_cmp);
 }
 
 static struct zk_node *zk_tree_search_nolock(const struct node_id *nid)
 {
-	struct rb_node *n = zk_node_root.rb_node;
-	struct zk_node *t;
+	struct zk_node key = { .node.nid = *nid };
 
-	while (n) {
-		int cmp;
-
-		t = rb_entry(n, struct zk_node, rb);
-		cmp = node_id_cmp(nid, &t->node.nid);
-
-		if (cmp < 0)
-			n = n->rb_left;
-		else if (cmp > 0)
-			n = n->rb_right;
-		else
-			return t; /* found it */
-	}
-	return NULL;
+	return rb_search(&zk_node_root, &key, rb, zk_node_cmp);
 }
 
 static inline struct zk_node *zk_tree_search(const struct node_id *nid)
@@ -134,7 +138,7 @@ static inline struct zk_node *zk_tree_search(const struct node_id *nid)
 
 	sd_read_lock(&zk_tree_lock);
 	n = zk_tree_search_nolock(nid);
-	sd_unlock(&zk_tree_lock);
+	sd_rw_unlock(&zk_tree_lock);
 	return n;
 }
 
@@ -146,6 +150,7 @@ static struct zk_node this_node;
 	switch (rc) {							\
 	case ZNONODE:							\
 	case ZNODEEXISTS:						\
+	case ZNOTEMPTY:							\
 		break;							\
 	case ZINVALIDSTATE:						\
 	case ZSESSIONEXPIRED:						\
@@ -295,6 +300,156 @@ static inline ZOOAPI int zk_get_children(const char *path,
 	return rc;
 }
 
+/*
+ * All the operations of the lock table is protected by
+ * cluster_lock->id_lock so we don't need to add lock here
+ */
+
+static int lock_table_lookup_wakeup(uint64_t lock_id)
+{
+	uint64_t hval = sd_hash_64(lock_id) % HASH_BUCKET_NR;
+	int res = -1;
+	struct hlist_node *iter;
+	struct cluster_lock *lock;
+
+	sd_mutex_lock(table_locks + hval);
+	hlist_for_each_entry(lock, iter, cluster_locks_table + hval, hnode) {
+		if (lock->id == lock_id) {
+			sd_cond_signal(&lock->wait_wakeup);
+			res = 0;
+			break;
+		}
+	}
+	sd_mutex_unlock(table_locks + hval);
+	return res;
+}
+
+static struct cluster_lock *lock_table_lookup_acquire(uint64_t lock_id)
+{
+	uint64_t hval = sd_hash_64(lock_id) % HASH_BUCKET_NR;
+	int rc;
+	struct hlist_node *iter;
+	struct cluster_lock *lock, *ret_lock = NULL;
+	char path[MAX_NODE_STR_LEN];
+
+	sd_mutex_lock(table_locks + hval);
+	hlist_for_each_entry(lock, iter, cluster_locks_table + hval, hnode) {
+		if (lock->id == lock_id) {
+			ret_lock = lock;
+			ret_lock->ref++;
+			break;
+		}
+	}
+
+	if (!ret_lock) {
+		/* create lock and add it to hash table */
+		ret_lock = xzalloc(sizeof(*ret_lock));
+		ret_lock->id = lock_id;
+		ret_lock->ref = 1;
+		snprintf(path, MAX_NODE_STR_LEN, LOCK_ZNODE "/%"PRIu64,
+			 ret_lock->id);
+		rc = zk_init_node(path);
+		if (rc)
+			panic("Failed to init node %s", path);
+
+		sd_cond_init(&ret_lock->wait_wakeup);
+		sd_init_mutex(&ret_lock->wait_mutex);
+		sd_init_mutex(&ret_lock->id_lock);
+
+		hlist_add_head(&(ret_lock->hnode), cluster_locks_table + hval);
+	}
+	sd_mutex_unlock(table_locks + hval);
+
+	/*
+	 * if many threads use locks with same id, we should use
+	 * ->id_lock to avoid the only zookeeper handler to
+	 * create many seq-ephemeral files.
+	 */
+	sd_mutex_lock(&ret_lock->id_lock);
+	return ret_lock;
+}
+
+static void lock_table_lookup_release(uint64_t lock_id)
+{
+	uint64_t hval = sd_hash_64(lock_id) % HASH_BUCKET_NR;
+	int rc;
+	struct hlist_node *iter;
+	struct cluster_lock *lock;
+	char path[MAX_NODE_STR_LEN];
+
+	sd_mutex_lock(table_locks + hval);
+	hlist_for_each_entry(lock, iter, cluster_locks_table + hval, hnode) {
+		if (lock->id != lock_id)
+			continue;
+		while (true) {
+			rc = zk_delete_node(lock->lock_path, -1);
+			if (rc == ZOK || rc == ZNONODE) {
+				sd_debug("delete path: %s ok", lock->lock_path);
+				break;
+			}
+			sd_err("Failed to delete path: %s %s", lock->lock_path,
+			       zerror(rc));
+			zk_wait();
+		}
+		lock->lock_path[0] = '\0';
+		sd_mutex_unlock(&lock->id_lock);
+		lock->ref--;
+		if (!lock->ref) {
+			hlist_del(iter);
+			/* free all resource used by this lock */
+			sd_destroy_mutex(&lock->id_lock);
+			sd_destroy_mutex(&lock->wait_mutex);
+			sd_destroy_cond(&lock->wait_wakeup);
+			snprintf(path, MAX_NODE_STR_LEN, LOCK_ZNODE "/%"PRIu64,
+				 lock->id);
+			/*
+			 * If deletion of directory 'lock_id' fail, we only get
+			 * a * empty directory in zookeeper. That's unharmful
+			 * so we don't need to retry it.
+			 */
+			rc = zk_delete_node(path, -1);
+			if (rc != ZOK)
+				sd_err("Failed to delete path: %s %s", path,
+				      zerror(rc));
+			free(lock);
+		}
+		break;
+	}
+	sd_mutex_unlock(table_locks + hval);
+}
+
+/*
+ * If this node leave the cluster, we need to delete the znode which created
+ * for distributed lock. Otherwise, the lock will never be released.
+ */
+static void lock_table_remove_znodes(void)
+{
+	uint64_t hval;
+	int rc;
+	struct hlist_node *iter;
+	struct cluster_lock *lock;
+
+	for (hval = 0; hval < HASH_BUCKET_NR; hval++) {
+		sd_mutex_lock(table_locks + hval);
+		hlist_for_each_entry(lock, iter, cluster_locks_table + hval,
+				     hnode) {
+			while (true) {
+				rc = zk_delete_node(lock->lock_path, -1);
+				if (rc == ZOK || rc == ZNONODE) {
+					sd_debug("delete path: %s ok",
+						 lock->lock_path);
+					break;
+				}
+				sd_err("Failed to delete path: %s %s",
+				       lock->lock_path,
+				       zerror(rc));
+				zk_wait();
+			}
+		}
+		sd_mutex_unlock(table_locks + hval);
+	}
+}
+
 /* ZooKeeper-based queue give us an totally ordered events */
 static int efd;
 static int32_t queue_pos;
@@ -403,12 +558,14 @@ static inline void *zk_event_sd_nodes(struct zk_event *ev)
 static int push_join_response(struct zk_event *ev)
 {
 	char path[MAX_NODE_STR_LEN];
+	struct sd_node *n, *np = zk_event_sd_nodes(ev);
 	int len;
 
 	ev->type = EVENT_ACCEPT;
 	ev->nr_nodes = nr_sd_nodes;
-	memcpy(zk_event_sd_nodes(ev), sd_nodes,
-	       nr_sd_nodes * sizeof(struct sd_node));
+	rb_for_each_entry(n, &sd_node_root, rb) {
+		memcpy(np++, n, sizeof(struct sd_node));
+	}
 	queue_pos--;
 
 	len = offsetof(typeof(*ev), buf) + ev->buf_len;
@@ -448,48 +605,38 @@ static inline void zk_tree_add(struct zk_node *node)
 	 * Even node list will be built later, we need this because in master
 	 * transfer case, we need this information to destroy the tree.
 	 */
-	sd_nodes[nr_sd_nodes++] = zk->node;
+	rb_insert(&sd_node_root, &zk->node, rb, node_cmp);
+	nr_sd_nodes++;
 out:
-	sd_unlock(&zk_tree_lock);
-}
-
-static inline void zk_tree_del_nolock(struct zk_node *node)
-{
-	rb_erase(&node->rb, &zk_node_root);
-	free(node);
+	sd_rw_unlock(&zk_tree_lock);
 }
 
 static inline void zk_tree_del(struct zk_node *node)
 {
 	sd_write_lock(&zk_tree_lock);
-	zk_tree_del_nolock(node);
-	sd_unlock(&zk_tree_lock);
+	rb_erase(&node->rb, &zk_node_root);
+	free(node);
+	sd_rw_unlock(&zk_tree_lock);
 }
 
 static inline void zk_tree_destroy(void)
 {
-	struct zk_node *zk;
-	int i;
-
 	sd_write_lock(&zk_tree_lock);
-	for (i = 0; i < nr_sd_nodes; i++) {
-		zk = zk_tree_search_nolock(&sd_nodes[i].nid);
-		if (zk)
-			zk_tree_del_nolock(zk);
-	}
-	sd_unlock(&zk_tree_lock);
+	rb_destroy(&zk_node_root, struct zk_node, rb);
+	sd_rw_unlock(&zk_tree_lock);
 }
 
 static inline void build_node_list(void)
 {
-	struct rb_node *n;
 	struct zk_node *zk;
 
 	nr_sd_nodes = 0;
-	for (n = rb_first(&zk_node_root); n; n = rb_next(n)) {
-		zk = rb_entry(n, struct zk_node, rb);
-		sd_nodes[nr_sd_nodes++] = zk->node;
+	INIT_RB_ROOT(&sd_node_root);
+	rb_for_each_entry(zk, &zk_node_root, rb) {
+		rb_insert(&sd_node_root, &zk->node, rb, node_cmp);
+		nr_sd_nodes++;
 	}
+
 	sd_debug("nr_sd_nodes:%zu", nr_sd_nodes);
 }
 
@@ -506,12 +653,15 @@ static int zk_queue_init(void)
 static uint64_t get_uniq_id(void)
 {
 	static int seq;
-	uint64_t id, n = uatomic_add_return(&seq, 1);
+	struct {
+		uint64_t n;
+		struct zk_node node;
+	} id = {
+		.n = uatomic_add_return(&seq, 1),
+		.node = this_node,
+	};
 
-	id = fnv_64a_buf(&this_node, sizeof(this_node), FNV1A_64_INIT);
-	id = fnv_64a_buf(&n, sizeof(n), id);
-
-	return id;
+	return sd_hash(&id, sizeof(id));
 }
 
 static int add_event(enum zk_event_type type, struct zk_node *znode, void *buf,
@@ -540,6 +690,7 @@ static void zk_watcher(zhandle_t *zh, int type, int state, const char *path,
 {
 	struct zk_node znode;
 	char str[MAX_NODE_STR_LEN], *p;
+	uint64_t lock_id;
 	int ret;
 
 	if (type == ZOO_SESSION_EVENT && state == ZOO_EXPIRED_SESSION_STATE) {
@@ -552,7 +703,7 @@ static void zk_watcher(zhandle_t *zh, int type, int state, const char *path,
 	}
 
 /* CREATED_EVENT 1, DELETED_EVENT 2, CHANGED_EVENT 3, CHILD_EVENT 4 */
-	sd_debug("path:%s, type:%d", path, type);
+	sd_debug("path:%s, type:%d, state:%d", path, type, state);
 	if (type == ZOO_CREATED_EVENT || type == ZOO_CHANGED_EVENT) {
 		ret = sscanf(path, MEMBER_ZNODE "/%s", str);
 		if (ret == 1)
@@ -562,6 +713,16 @@ static void zk_watcher(zhandle_t *zh, int type, int state, const char *path,
 	} else if (type == ZOO_DELETED_EVENT) {
 		struct zk_node *n;
 
+		/* process distributed lock */
+		ret = sscanf(path, LOCK_ZNODE "/%"PRIu64"/%s", &lock_id, str);
+		if (ret == 2) {
+			ret = lock_table_lookup_wakeup(lock_id);
+			if (ret)
+				sd_debug("release lock %"PRIu64" %s",
+					 lock_id, str);
+			return;
+		}
+
 		ret = sscanf(path, MASTER_ZNONE "/%s", str);
 		if (ret == 1) {
 			zk_compete_master();
@@ -579,7 +740,7 @@ static void zk_watcher(zhandle_t *zh, int type, int state, const char *path,
 		n = zk_tree_search_nolock(&znode.node.nid);
 		if (n)
 			n->gone = true;
-		sd_unlock(&zk_tree_lock);
+		sd_rw_unlock(&zk_tree_lock);
 		if (n)
 			add_event(EVENT_LEAVE, &znode, NULL, 0);
 	}
@@ -588,20 +749,24 @@ static void zk_watcher(zhandle_t *zh, int type, int state, const char *path,
 /*
  * We placehold the enough space to piggyback the nodes information on join
  * response message so that every node can see the same membership view.
+ *
+ * We have to preallocate enough space and set msg_len as
+ * sizeof(struct cluster_info) because of piggyback.
  */
-static int add_join_event(void *msg, size_t msg_len)
+static int add_join_event(void *msg, size_t msglen)
 {
 	struct zk_event ev;
-	size_t len = msg_len + sizeof(struct sd_node) * SD_MAX_NODES;
+	size_t len = msglen + sizeof(struct sd_node) * SD_MAX_NODES;
 
-	assert(len <= SD_MAX_EVENT_BUF_SIZE);
+	if (unlikely((offsetof(struct zk_event, buf) + len) > ZK_MAX_BUF_SIZE))
+		panic("Zookeeper can't send message more than 1M");
 	ev.id = get_uniq_id();
 	ev.type = EVENT_JOIN;
 	ev.sender = this_node;
-	ev.msg_len = msg_len;
+	ev.msg_len = msglen;
 	ev.buf_len = len;
 	if (msg)
-		memcpy(ev.buf, msg, msg_len);
+		memcpy(ev.buf, msg, msglen);
 	return zk_queue_push(&ev);
 }
 
@@ -610,9 +775,17 @@ static int zk_get_least_seq(const char *parent, char *least_seq_path,
 {
 	char path[MAX_NODE_STR_LEN], *p, *tmp;
 	struct String_vector strs;
-	int rc, least_seq = INT_MAX , seq;
+	int rc, least_seq, seq;
 
 	while (true) {
+		/*
+		 * If first loop fail, the least_seq may be a very small number
+		 * which had been deleted in zookeeper, the new create file will
+		 * be all larger than it and it will cause dead loop.
+		 * Therefore we need to set least_seq to INT_MAX in every loop.
+		 */
+		least_seq = INT_MAX;
+
 		RETURN_IF_ERROR(zk_get_children(parent, &strs), "");
 
 		FOR_EACH_ZNODE(parent, path, &strs) {
@@ -782,7 +955,7 @@ success:
 	uatomic_set_true(&is_master);
 	sd_debug("success");
 out_unlock:
-	sd_unlock(&zk_compete_master_lock);
+	sd_rw_unlock(&zk_compete_master_lock);
 }
 
 static int zk_join(const struct sd_node *myself,
@@ -796,7 +969,8 @@ static int zk_join(const struct sd_node *myself,
 	snprintf(path, sizeof(path), MEMBER_ZNODE "/%s", node_to_str(myself));
 	rc = zk_node_exists(path);
 	if (rc == ZOK) {
-		sd_err("Previous zookeeper session exist, shoot myself.");
+		sd_err("Previous zookeeper session exist, shoot myself. Please "
+		       "wait for %d seconds to join me again.", zk_timeout);
 		exit(1);
 	}
 
@@ -816,6 +990,7 @@ static int zk_leave(void)
 	snprintf(path, sizeof(path), MEMBER_ZNODE"/%s",
 		 node_to_str(&this_node.node));
 	add_event(EVENT_LEAVE, &this_node, NULL, 0);
+	lock_table_remove_znodes();
 	zk_delete_node(path, -1);
 	return 0;
 }
@@ -844,7 +1019,7 @@ static void zk_handle_join(struct zk_event *ev)
 		return;
 	}
 
-	sd_join_handler(&ev->sender.node, sd_nodes, nr_sd_nodes, ev->buf);
+	sd_join_handler(&ev->sender.node, &sd_node_root, nr_sd_nodes, ev->buf);
 	push_join_response(ev);
 
 	sd_debug("I'm the master now");
@@ -908,7 +1083,8 @@ static void zk_handle_accept(struct zk_event *ev)
 	zk_tree_add(&ev->sender);
 
 	build_node_list();
-	sd_accept_handler(&ev->sender.node, sd_nodes, nr_sd_nodes, ev->buf);
+	sd_accept_handler(&ev->sender.node, &sd_node_root, nr_sd_nodes,
+			  ev->buf);
 }
 
 static void kick_block_event(void)
@@ -924,9 +1100,9 @@ static void kick_block_event(void)
 
 static void block_event_list_del(struct zk_node *n)
 {
-	struct zk_node *ev, *t;
+	struct zk_node *ev;
 
-	list_for_each_entry_safe(ev, t, &zk_block_list, list) {
+	list_for_each_entry(ev, &zk_block_list, list) {
 		if (node_eq(&ev->node, &n->node)) {
 			list_del(&ev->list);
 			free(ev);
@@ -946,7 +1122,7 @@ static void zk_handle_leave(struct zk_event *ev)
 	block_event_list_del(n);
 	zk_tree_del(n);
 	build_node_list();
-	sd_leave_handler(&ev->sender.node, sd_nodes, nr_sd_nodes);
+	sd_leave_handler(&ev->sender.node, &sd_node_root, nr_sd_nodes);
 }
 
 static void zk_handle_block(struct zk_event *ev)
@@ -996,7 +1172,7 @@ static void zk_handle_update_node(struct zk_event *ev)
 	assert(t);
 	t->node = *snode;
 	build_node_list();
-	sd_unlock(&zk_tree_lock);
+	sd_rw_unlock(&zk_tree_lock);
 	sd_update_node_handler(snode);
 }
 
@@ -1027,9 +1203,9 @@ static inline void handle_session_expire(void)
 	INIT_RB_ROOT(&zk_node_root);
 	INIT_LIST_HEAD(&zk_block_list);
 	nr_sd_nodes = 0;
+	INIT_RB_ROOT(&sd_node_root);
 	first_push = true;
 	joined = false;
-	memset(sd_nodes, 0, sizeof(struct sd_node) * SD_MAX_NODES);
 
 	while (sd_reconnect_handler()) {
 		sd_err("failed to reconnect. sleep and retry...");
@@ -1087,10 +1263,98 @@ kick_block_event:
 	kick_block_event();
 }
 
+/*
+ * This operation will create a seq-ephemeral znode in lock directory
+ * of zookeeper (use lock-id as dir name). The smallest file path in
+ * this directory wil be the owner of the lock; the other threads will
+ * wait on a struct sd_cond (cluster_lock->wait_wakeup)
+ */
+static void zk_lock(uint64_t lock_id)
+{
+	int rc, len = MAX_NODE_STR_LEN;
+	char *my_path;
+	char parent[MAX_NODE_STR_LEN];
+	char parent_node[MAX_NODE_STR_LEN];
+	char lowest_seq_path[MAX_NODE_STR_LEN];
+	char owner_name[MAX_NODE_STR_LEN];
+	struct cluster_lock *cluster_lock;
+
+	cluster_lock = lock_table_lookup_acquire(lock_id);
+
+	my_path = cluster_lock->lock_path;
+
+	snprintf(parent, MAX_NODE_STR_LEN, LOCK_ZNODE "/%"PRIu64"/",
+		 cluster_lock->id);
+	/*
+	 * It need using path without end of '/' to create node of lock_id in
+	 * zookeeper's API, so we use 'parent_node'.
+	 */
+	snprintf(parent_node, MAX_NODE_STR_LEN, LOCK_ZNODE "/%"PRIu64,
+		 cluster_lock->id);
+create_seq_node:
+	/* compete owner of lock is just like zk_compete_master() */
+	while (true) {
+		rc = zk_create_seq_node(parent, node_to_str(&this_node.node),
+					MAX_NODE_STR_LEN, my_path,
+					MAX_NODE_STR_LEN, true);
+		if (rc == ZOK)
+			break;
+		if (rc == ZNONODE) {
+			zk_init_node(parent_node);
+			/*
+			 * We don't need to check the return code of
+			 * zk_init_node() because the routine must stay in loop
+			 * if it doesn't take the lock, no matter what kind of
+			 * error happed.
+			 */
+			continue;
+		}
+		sd_err("failed to create path:%s, %s", my_path, zerror(rc));
+		zk_wait();
+	}
+	sd_debug("create path %s success", my_path);
+
+	/* create node ok now */
+	while (true) {
+		rc = zk_get_least_seq(parent_node, lowest_seq_path,
+				      MAX_NODE_STR_LEN, owner_name, &len);
+		/* may be expired */
+		if (rc == ZNONODE) {
+			sd_debug("Recreate seq node");
+			goto create_seq_node;
+		}
+		/* I got the lock */
+		if (!strncmp(lowest_seq_path, my_path, strlen(my_path))) {
+			sd_debug("I am master now. %s", lowest_seq_path);
+			return;
+		}
+
+		/* I failed to get the lock */
+		rc = zk_node_exists(lowest_seq_path);
+		if (rc == ZOK) {
+			sd_debug("call zoo_exists success %s", lowest_seq_path);
+			/* Use wait_timeout to avoid missing wakeup signal */
+			sd_cond_wait_timeout(&cluster_lock->wait_wakeup,
+					     &cluster_lock->wait_mutex,
+					     WAIT_TIME * 2);
+		} else {
+			sd_debug("failed to call zoo_exists %s", zerror(rc));
+			if (rc != ZNONODE)
+				zk_wait();
+		}
+	}
+}
+
+static void zk_unlock(uint64_t lock_id)
+{
+	lock_table_lookup_release(lock_id);
+	sd_debug("unlock %"PRIu64, lock_id);
+}
+
 static int zk_init(const char *option)
 {
 	char *hosts, *to, *p;
-	int ret, timeout = SESSION_TIMEOUT;
+	int ret;
 
 	if (!option) {
 		sd_err("You must specify zookeeper servers.");
@@ -1099,7 +1363,7 @@ static int zk_init(const char *option)
 
 	hosts = strtok((char *)option, "=");
 	if ((to = strtok(NULL, "="))) {
-		if (sscanf(to, "%u", &timeout) != 1) {
+		if (sscanf(to, "%u", &zk_timeout) != 1) {
 			sd_err("Invalid paramter for timeout");
 			return -1;
 		}
@@ -1107,8 +1371,8 @@ static int zk_init(const char *option)
 		*--p = '\0';
 	}
 	sd_debug("version %d.%d.%d, address %s, timeout %d", ZOO_MAJOR_VERSION,
-		 ZOO_MINOR_VERSION, ZOO_PATCH_VERSION, hosts, timeout);
-	zhandle = zookeeper_init(hosts, zk_watcher, timeout, NULL, NULL, 0);
+		 ZOO_MINOR_VERSION, ZOO_PATCH_VERSION, hosts, zk_timeout);
+	zhandle = zookeeper_init(hosts, zk_watcher, zk_timeout, NULL, NULL, 0);
 	if (!zhandle) {
 		sd_err("failed to connect to zk server %s", option);
 		return -1;
@@ -1131,6 +1395,20 @@ static int zk_init(const char *option)
 		return -1;
 	}
 
+	/* init distributed lock structures */
+	cluster_locks_table = xzalloc(sizeof(struct list_head) *
+				      HASH_BUCKET_NR);
+	for (uint64_t i = 0; i < HASH_BUCKET_NR; i++) {
+		INIT_HLIST_HEAD(cluster_locks_table + i);
+		sd_init_mutex(table_locks + i);
+	}
+
+	ret = zk_init_node(LOCK_ZNODE);
+	if (ret != ZOK) {
+		sd_err("Failed to create %s %s", LOCK_ZNODE, zerror(ret));
+		free(cluster_locks_table);
+		return -1;
+	}
 	return 0;
 }
 
@@ -1151,7 +1429,9 @@ static struct cluster_driver cdrv_zookeeper = {
 	.notify     = zk_notify,
 	.block      = zk_block,
 	.unblock    = zk_unblock,
-	.update_node = zk_update_node,
+	.lock         = zk_lock,
+	.unlock       = zk_unlock,
+	.update_node  = zk_update_node,
 	.get_local_addr = get_local_addr,
 };
 
