@@ -13,45 +13,45 @@
 
 struct node {
 	struct sd_node ent;
-	struct list_head list;
+	struct list_node list;
 };
 
 struct get_vdis_work {
 	struct work work;
 	DECLARE_BITMAP(vdi_inuse, SD_NR_VDIS);
 	struct sd_node joined;
-	size_t nr_members;
-	struct sd_node members[];
+	struct rb_root nroot;
 };
 
-static pthread_mutex_t wait_vdis_lock = PTHREAD_MUTEX_INITIALIZER;
-static pthread_cond_t wait_vdis_cond = PTHREAD_COND_INITIALIZER;
+static struct sd_mutex wait_vdis_lock = SD_MUTEX_INITIALIZER;
+static struct sd_cond wait_vdis_cond = SD_COND_INITIALIZER;
 static refcnt_t nr_get_vdis_works;
 
 static main_thread(struct vnode_info *) current_vnode_info;
 static main_thread(struct list_head *) pending_block_list;
 static main_thread(struct list_head *) pending_notify_list;
 
-static int get_zones_nr_from(const struct sd_node *nodes, int nr_nodes)
+static int get_zones_nr_from(struct rb_root *nroot)
 {
-	int nr_zones = 0, i, j;
+	int nr_zones = 0, j;
 	uint32_t zones[SD_MAX_COPIES];
+	struct sd_node *n;
 
-	for (i = 0; i < nr_nodes; i++) {
+	rb_for_each_entry(n, nroot, rb) {
 		/*
 		 * Only count zones that actually store data, pure gateways
 		 * don't contribute to the redundancy level.
 		 */
-		if (!nodes[i].nr_vnodes)
+		if (!n->nr_vnodes)
 			continue;
 
 		for (j = 0; j < nr_zones; j++) {
-			if (nodes[i].zone == zones[j])
+			if (n->zone == zones[j])
 				break;
 		}
 
 		if (j == nr_zones) {
-			zones[nr_zones] = nodes[i].zone;
+			zones[nr_zones] = n->zone;
 			if (++nr_zones == ARRAY_SIZE(zones))
 				break;
 		}
@@ -60,15 +60,6 @@ static int get_zones_nr_from(const struct sd_node *nodes, int nr_nodes)
 	return nr_zones;
 }
 
-static int get_node_idx(struct vnode_info *vnode_info, struct sd_node *ent)
-{
-	ent = xbsearch(ent, vnode_info->nodes, vnode_info->nr_nodes, node_cmp);
-	if (!ent)
-		return -1;
-
-	return ent - vnode_info->nodes;
-}
-
 /*
  * Grab an additional reference to the passed in vnode info.
  *
@@ -101,27 +92,62 @@ main_fn struct vnode_info *get_vnode_info(void)
 void put_vnode_info(struct vnode_info *vnode_info)
 {
 	if (vnode_info) {
-		if (refcount_dec(&vnode_info->refcnt) == 0)
+		if (refcount_dec(&vnode_info->refcnt) == 0) {
+			rb_destroy(&vnode_info->vroot, struct sd_vnode, rb);
+			rb_destroy(&vnode_info->nroot, struct sd_node, rb);
 			free(vnode_info);
+		}
 	}
 }
 
-struct vnode_info *alloc_vnode_info(const struct sd_node *nodes,
-				    size_t nr_nodes)
+static void recalculate_vnodes(struct rb_root *nroot)
+{
+	int nr_non_gateway_nodes = 0;
+	uint64_t avg_size = 0;
+	struct sd_node *n;
+	float factor;
+
+	rb_for_each_entry(n, nroot, rb) {
+		if (n->space) {
+			avg_size += n->space;
+			nr_non_gateway_nodes++;
+		}
+	}
+
+	if (!nr_non_gateway_nodes)
+		return;
+
+	avg_size /= nr_non_gateway_nodes;
+
+	rb_for_each_entry(n, nroot, rb) {
+		factor = (float)n->space / (float)avg_size;
+		n->nr_vnodes = rintf(SD_DEFAULT_VNODES * factor);
+		sd_debug("node %s has %d vnodes, free space %" PRIu64,
+			 node_to_str(n), n->nr_vnodes, n->space);
+	}
+}
+
+struct vnode_info *alloc_vnode_info(const struct rb_root *nroot)
 {
 	struct vnode_info *vnode_info;
+	struct sd_node *n;
 
 	vnode_info = xzalloc(sizeof(*vnode_info));
 
-	vnode_info->nr_nodes = nr_nodes;
-	memcpy(vnode_info->nodes, nodes, sizeof(*nodes) * nr_nodes);
-	xqsort(vnode_info->nodes, nr_nodes, node_cmp);
+	INIT_RB_ROOT(&vnode_info->vroot);
+	INIT_RB_ROOT(&vnode_info->nroot);
+	rb_for_each_entry(n, nroot, rb) {
+		struct sd_node *new = xmalloc(sizeof(*new));
+		*new = *n;
+		if (unlikely(rb_insert(&vnode_info->nroot, new, rb, node_cmp)))
+			panic("node hash collision");
+		vnode_info->nr_nodes++;
+	}
 
-	recalculate_vnodes(vnode_info->nodes, nr_nodes);
+	recalculate_vnodes(&vnode_info->nroot);
 
-	vnode_info->nr_vnodes = nodes_to_vnodes(vnode_info->nodes, nr_nodes,
-						vnode_info->vnodes);
-	vnode_info->nr_zones = get_zones_nr_from(nodes, nr_nodes);
+	nodes_to_vnodes(&vnode_info->nroot, &vnode_info->vroot);
+	vnode_info->nr_zones = get_zones_nr_from(&vnode_info->nroot);
 	refcount_set(&vnode_info->refcnt, 1);
 	return vnode_info;
 }
@@ -130,6 +156,7 @@ struct vnode_info *get_vnode_info_epoch(uint32_t epoch,
 					struct vnode_info *cur_vinfo)
 {
 	struct sd_node nodes[SD_MAX_NODES];
+	struct rb_root nroot = RB_ROOT;
 	int nr_nodes;
 
 	nr_nodes = epoch_log_read(epoch, nodes, sizeof(nodes));
@@ -139,26 +166,27 @@ struct vnode_info *get_vnode_info_epoch(uint32_t epoch,
 		if (nr_nodes == 0)
 			return NULL;
 	}
+	for (int i = 0; i < nr_nodes; i++)
+		rb_insert(&nroot, &nodes[i], rb, node_cmp);
 
-	return alloc_vnode_info(nodes, nr_nodes);
+	return alloc_vnode_info(&nroot);
 }
 
 int local_get_node_list(const struct sd_req *req, struct sd_rsp *rsp,
-			       void *data)
+			void *data)
 {
 	int nr_nodes;
-	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
+	struct vnode_info *cur_vinfo = get_vnode_info();
 
 	if (cur_vinfo) {
 		nr_nodes = cur_vinfo->nr_nodes;
-		memcpy(data, cur_vinfo->nodes,
-			sizeof(struct sd_node) * nr_nodes);
+		nodes_to_buffer(&cur_vinfo->nroot, data);
 		rsp->data_length = nr_nodes * sizeof(struct sd_node);
 		rsp->node.nr_nodes = nr_nodes;
-		rsp->node.local_idx = get_node_idx(cur_vinfo, &sys->this_node);
+
+		put_vnode_info(cur_vinfo);
 	} else {
 		rsp->node.nr_nodes = 0;
-		rsp->node.local_idx = 0;
 	}
 
 	return SD_RES_SUCCESS;
@@ -306,27 +334,16 @@ error:
 	put_request(req);
 }
 
-static inline int get_nodes_nr_from(struct list_head *l)
-{
-	struct node *node;
-	int nr = 0;
-	list_for_each_entry(node, l, list) {
-		nr++;
-	}
-	return nr;
-}
-
 int epoch_log_read_remote(uint32_t epoch, struct sd_node *nodes, int len,
 			  time_t *timestamp, struct vnode_info *vinfo)
 {
-	int i, nr, ret;
 	char buf[SD_MAX_NODES * sizeof(struct sd_node) + sizeof(time_t)];
+	const struct sd_node *node;
+	int ret;
 
-	nr = vinfo->nr_nodes;
-	for (i = 0; i < nr; i++) {
+	rb_for_each_entry(node, &vinfo->nroot, rb) {
 		struct sd_req hdr;
 		struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
-		const struct sd_node *node = vinfo->nodes + i;
 		int nodes_len;
 
 		if (node_is_local(node))
@@ -340,10 +357,10 @@ int epoch_log_read_remote(uint32_t epoch, struct sd_node *nodes, int len,
 		if (ret != SD_RES_SUCCESS)
 			continue;
 
-		nodes_len = rsp->data_length - sizeof(timestamp);
+		nodes_len = rsp->data_length - sizeof(*timestamp);
 		memcpy((void *)nodes, buf, nodes_len);
 		if (timestamp)
-			memcpy(timestamp, buf + nodes_len, sizeof(timestamp));
+			memcpy(timestamp, buf + nodes_len, sizeof(*timestamp));
 
 		return nodes_len / sizeof(struct sd_node);
 	}
@@ -377,13 +394,13 @@ static bool cluster_ctime_check(const struct cluster_info *cinfo)
  */
 static bool enough_nodes_gathered(struct cluster_info *cinfo,
 				  const struct sd_node *joining,
-				  const struct sd_node *nodes,
+				  const struct rb_root *nroot,
 				  size_t nr_nodes)
 {
 	for (int i = 0; i < cinfo->nr_nodes; i++) {
 		const struct sd_node *key = cinfo->nodes + i, *n;
 
-		n = xlfind(key, nodes, nr_nodes, node_cmp);
+		n = rb_search(nroot, key, rb, node_cmp);
 		if (n == NULL && !node_eq(key, joining)) {
 			sd_debug("%s doesn't join yet", node_to_str(key));
 			return false;
@@ -395,8 +412,20 @@ static bool enough_nodes_gathered(struct cluster_info *cinfo,
 	return true;
 }
 
+/*
+ * We have to use memcpy beause some cluster drivers like corosync can't support
+ * to send the whole cluster_info structure.
+ */
+static void cluster_info_copy(struct cluster_info *dst,
+			      const struct cluster_info *src)
+{
+	int len = offsetof(struct cluster_info, nodes) +
+		src->nr_nodes * sizeof(struct sd_node);
+	memcpy(dst, src, len);
+}
+
 static enum sd_status cluster_wait_check(const struct sd_node *joining,
-					 const struct sd_node *nodes,
+					 const struct rb_root *nroot,
 					 size_t nr_nodes,
 					 struct cluster_info *cinfo)
 {
@@ -408,7 +437,7 @@ static enum sd_status cluster_wait_check(const struct sd_node *joining,
 	if (cinfo->epoch > sys->cinfo.epoch) {
 		sd_debug("joining node has a larger epoch, %" PRIu32 ", %"
 			 PRIu32, cinfo->epoch, sys->cinfo.epoch);
-		sys->cinfo = *cinfo;
+		cluster_info_copy(&sys->cinfo, cinfo);
 	}
 
 	/*
@@ -416,7 +445,7 @@ static enum sd_status cluster_wait_check(const struct sd_node *joining,
 	 * node list, we can set the cluster live now.
 	 */
 	if (sys->cinfo.epoch > 0 &&
-	    enough_nodes_gathered(&sys->cinfo, joining, nodes, nr_nodes))
+	    enough_nodes_gathered(&sys->cinfo, joining, nroot, nr_nodes))
 		return SD_STATUS_OK;
 
 	return sys->cinfo.status;
@@ -446,7 +475,8 @@ static int get_vdis_from(struct sd_node *node)
 	count = rsp->data_length / sizeof(*vs);
 	for (i = 0; i < count; i++) {
 		atomic_set_bit(vs[i].vid, sys->vdi_inuse);
-		add_vdi_state(vs[i].vid, vs[i].nr_copies, vs[i].snapshot);
+		add_vdi_state(vs[i].vid, vs[i].nr_copies, vs[i].snapshot,
+			      vs[i].copy_policy);
 	}
 out:
 	free(vs);
@@ -457,7 +487,8 @@ static void do_get_vdis(struct work *work)
 {
 	struct get_vdis_work *w =
 		container_of(work, struct get_vdis_work, work);
-	int i, ret;
+	struct sd_node *n;
+	int ret;
 
 	if (!node_is_local(&w->joined)) {
 		sd_debug("try to get vdi bitmap from %s",
@@ -469,18 +500,17 @@ static void do_get_vdis(struct work *work)
 		return;
 	}
 
-	for (i = 0; i < w->nr_members; i++) {
+	rb_for_each_entry(n, &w->nroot, rb) {
 		/* We should not fetch vdi_bitmap and copy list from myself */
-		if (node_is_local(&w->members[i]))
+		if (node_is_local(n))
 			continue;
 
-		sd_debug("try to get vdi bitmap from %s",
-			 node_to_str(&w->members[i]));
-		ret = get_vdis_from(&w->members[i]);
+		sd_debug("try to get vdi bitmap from %s", node_to_str(n));
+		ret = get_vdis_from(n);
 		if (ret != SD_RES_SUCCESS) {
 			/* try to read from another node */
 			sd_alert("failed to get vdi bitmap from %s",
-				 node_to_str(&w->members[i]));
+				 node_to_str(n));
 			continue;
 		}
 
@@ -497,23 +527,25 @@ static void get_vdis_done(struct work *work)
 	struct get_vdis_work *w =
 		container_of(work, struct get_vdis_work, work);
 
-	pthread_mutex_lock(&wait_vdis_lock);
+	sd_mutex_lock(&wait_vdis_lock);
 	refcount_dec(&nr_get_vdis_works);
-	pthread_cond_broadcast(&wait_vdis_cond);
-	pthread_mutex_unlock(&wait_vdis_lock);
+	sd_cond_broadcast(&wait_vdis_cond);
+	sd_mutex_unlock(&wait_vdis_lock);
 
+	rb_destroy(&w->nroot, struct sd_node, rb);
 	free(w);
 }
 
 int inc_and_log_epoch(void)
 {
-	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
+	struct vnode_info *cur_vinfo = get_vnode_info();
 
 	if (cur_vinfo) {
 		/* update cluster info to the latest state */
 		sys->cinfo.nr_nodes = cur_vinfo->nr_nodes;
-		memcpy(sys->cinfo.nodes, cur_vinfo->nodes,
-		       sizeof(cur_vinfo->nodes[0]) * cur_vinfo->nr_nodes);
+		nodes_to_buffer(&cur_vinfo->nroot, sys->cinfo.nodes);
+
+		put_vnode_info(cur_vinfo);
 	} else
 		sys->cinfo.nr_nodes = 0;
 
@@ -524,16 +556,28 @@ int inc_and_log_epoch(void)
 }
 
 static struct vnode_info *alloc_old_vnode_info(const struct sd_node *joined,
-					       const struct sd_node *nodes,
-					       size_t nr_nodes)
+					       const struct rb_root *nroot)
 {
-	struct sd_node old_nodes[SD_MAX_NODES];
+	struct rb_root old_root = RB_ROOT;
+	struct sd_node *n;
+	struct vnode_info *old;
 
 	/* exclude the newly added one */
-	memcpy(old_nodes, nodes, sizeof(*nodes) * nr_nodes);
-	xlremove(joined, old_nodes, &nr_nodes, node_cmp);
+	rb_for_each_entry(n, nroot, rb) {
+		struct sd_node *new = xmalloc(sizeof(*new));
 
-	return alloc_vnode_info(old_nodes, nr_nodes);
+		*new = *n;
+		if (node_eq(joined, new)) {
+			free(new);
+			continue;
+		}
+		if (rb_insert(&old_root, new, rb, node_cmp))
+			panic("node hash collision");
+	}
+
+	old = alloc_vnode_info(&old_root);
+	rb_destroy(&old_root, struct sd_node, rb);
+	return old;
 }
 
 static void setup_backend_store(const struct cluster_info *cinfo)
@@ -565,22 +609,14 @@ static void setup_backend_store(const struct cluster_info *cinfo)
 	}
 }
 
-static void finish_join(const struct sd_node *nodes, size_t nr_nodes)
+static void get_vdis(const struct rb_root *nroot, const struct sd_node *joined)
 {
-	sockfd_cache_add_group(nodes, nr_nodes);
-}
-
-static void get_vdis(const struct sd_node *nodes, size_t nr_nodes,
-		     const struct sd_node *joined)
-{
-	int array_len = nr_nodes * sizeof(struct sd_node);
 	struct get_vdis_work *w;
 
-	w = xmalloc(sizeof(*w) + array_len);
+	w = xmalloc(sizeof(*w));
 	w->joined = *joined;
-	w->nr_members = nr_nodes;
-	memcpy(w->members, nodes, array_len);
-
+	INIT_RB_ROOT(&w->nroot);
+	rb_copy(nroot, struct sd_node, rb, &w->nroot, node_cmp);
 	refcount_inc(&nr_get_vdis_works);
 
 	w->work.fn = do_get_vdis;
@@ -592,43 +628,17 @@ void wait_get_vdis_done(void)
 {
 	sd_debug("waiting for vdi list");
 
-	pthread_mutex_lock(&wait_vdis_lock);
+	sd_mutex_lock(&wait_vdis_lock);
 	while (refcount_read(&nr_get_vdis_works) > 0)
-		pthread_cond_wait(&wait_vdis_cond, &wait_vdis_lock);
-	pthread_mutex_unlock(&wait_vdis_lock);
+		sd_cond_wait(&wait_vdis_cond, &wait_vdis_lock);
+	sd_mutex_unlock(&wait_vdis_lock);
 
 	sd_debug("vdi list ready");
 }
 
-void recalculate_vnodes(struct sd_node *nodes, int nr_nodes)
-{
-	int i, nr_non_gateway_nodes = 0;
-	uint64_t avg_size = 0;
-	float factor;
-
-	for (i = 0; i < nr_nodes; i++) {
-		if (nodes[i].space) {
-			avg_size += nodes[i].space;
-			nr_non_gateway_nodes++;
-		}
-	}
-
-	if (!nr_non_gateway_nodes)
-		return;
-
-	avg_size /= nr_non_gateway_nodes;
-
-	for (i = 0; i < nr_nodes; i++) {
-		factor = (float)nodes[i].space / (float)avg_size;
-		nodes[i].nr_vnodes = rintf(SD_DEFAULT_VNODES * factor);
-		sd_debug("node %d has %d vnodes, free space %" PRIu64,
-			 nodes[i].nid.port, nodes[i].nr_vnodes, nodes[i].space);
-	}
-}
-
 static void update_cluster_info(const struct cluster_info *cinfo,
 				const struct sd_node *joined,
-				const struct sd_node *nodes,
+				const struct rb_root *nroot,
 				size_t nr_nodes)
 {
 	struct vnode_info *old_vnode_info;
@@ -639,13 +649,19 @@ static void update_cluster_info(const struct cluster_info *cinfo,
 		setup_backend_store(cinfo);
 
 	if (node_is_local(joined))
-		finish_join(nodes, nr_nodes);
+		sockfd_cache_add_group(nroot);
+	sockfd_cache_add(&joined->nid);
 
+	/*
+	 * We need use main_thread_get() to obtain current_vnode_info. The
+	 * reference count of old_vnode_info is decremented at the last of this
+	 * function in order to release old_vnode_info. The counter part
+	 * of this dereference is alloc_vnode_info().
+	 */
 	old_vnode_info = main_thread_get(current_vnode_info);
-	main_thread_set(current_vnode_info,
-			  alloc_vnode_info(nodes, nr_nodes));
+	main_thread_set(current_vnode_info, alloc_vnode_info(nroot));
 
-	get_vdis(nodes, nr_nodes, joined);
+	get_vdis(nroot, joined);
 
 	if (cinfo->status == SD_STATUS_OK) {
 		if (!is_cluster_formatted())
@@ -658,22 +674,21 @@ static void update_cluster_info(const struct cluster_info *cinfo,
 				panic("cannot log current epoch %d",
 				      sys->cinfo.epoch);
 
-			if (!old_vnode_info) {
+			if (!old_vnode_info)
 				old_vnode_info = alloc_old_vnode_info(joined,
-						nodes, nr_nodes);
-			}
+								      nroot);
 
 			start_recovery(main_thread_get(current_vnode_info),
 				       old_vnode_info, true);
-		} else
+		} else if (!was_cluster_shutdowned()) {
 			start_recovery(main_thread_get(current_vnode_info),
 				       main_thread_get(current_vnode_info),
 				       false);
+		}
+		set_cluster_shutdown(false);
 	}
 
 	put_vnode_info(old_vnode_info);
-
-	sockfd_cache_add(&joined->nid);
 }
 
 /*
@@ -731,7 +746,7 @@ main_fn void sd_notify_handler(const struct sd_node *sender, void *data,
  * cluster must call this function and succeed in accept of the joining node.
  */
 main_fn bool sd_join_handler(const struct sd_node *joining,
-			     const struct sd_node *nodes, size_t nr_nodes,
+			     const struct rb_root *nroot, size_t nr_nodes,
 			     void *opaque)
 {
 	struct cluster_info *cinfo = opaque;
@@ -750,11 +765,11 @@ main_fn bool sd_join_handler(const struct sd_node *joining,
 	sd_debug("check %s, %d", node_to_str(joining), sys->cinfo.status);
 
 	if (sys->cinfo.status == SD_STATUS_WAIT)
-		status = cluster_wait_check(joining, nodes, nr_nodes, cinfo);
+		status = cluster_wait_check(joining, nroot, nr_nodes, cinfo);
 	else
 		status = sys->cinfo.status;
 
-	*cinfo = sys->cinfo;
+	cluster_info_copy(cinfo, &sys->cinfo);
 	cinfo->status = status;
 	cinfo->proto_ver = SD_SHEEP_PROTO_VER;
 
@@ -765,21 +780,22 @@ main_fn bool sd_join_handler(const struct sd_node *joining,
 	return true;
 }
 
-static int send_join_request(struct sd_node *ent)
+static int send_join_request(void)
 {
-	sd_info("%s", node_to_str(&sys->this_node));
+	struct sd_node *n = &sys->this_node;
 
-	return sys->cdrv->join(ent, &sys->cinfo, sizeof(sys->cinfo));
+	sd_info("%s", node_to_str(n));
+	return sys->cdrv->join(n, &sys->cinfo, sizeof(sys->cinfo));
 }
 
 static void requeue_cluster_request(void)
 {
-	struct request *req, *p;
+	struct request *req;
 	struct vdi_op_message *msg;
 	size_t size;
 
-	list_for_each_entry_safe(req, p, main_thread_get(pending_notify_list),
-				 pending_list) {
+	list_for_each_entry(req, main_thread_get(pending_notify_list),
+			    pending_list) {
 		/*
 		 * ->notify() was called and succeeded but after that
 		 * this node session-timeouted and sd_notify_handler
@@ -794,8 +810,8 @@ static void requeue_cluster_request(void)
 		free(msg);
 	}
 
-	list_for_each_entry_safe(req, p, main_thread_get(pending_block_list),
-				 pending_list) {
+	list_for_each_entry(req, main_thread_get(pending_block_list),
+			    pending_list) {
 		switch (req->status) {
 		case REQUEST_INIT:
 			/* this request has never been executed, re-queue it */
@@ -842,7 +858,7 @@ main_fn int sd_reconnect_handler(void)
 	sys->cinfo.status = SD_STATUS_WAIT;
 	if (sys->cdrv->init(sys->cdrv_option) != 0)
 		return -1;
-	if (send_join_request(&sys->this_node) != 0)
+	if (send_join_request() != 0)
 		return -1;
 	requeue_cluster_request();
 	return 0;
@@ -859,38 +875,44 @@ static bool cluster_join_check(const struct cluster_info *cinfo)
 	if (!cluster_ctime_check(cinfo))
 		return false;
 
-	if (cinfo->epoch == sys->cinfo.epoch &&
-	    memcmp(cinfo->nodes, sys->cinfo.nodes,
-		   sizeof(cinfo->nodes[0]) * cinfo->nr_nodes) != 0) {
-		sd_alert("epoch log entries does not match");
-		return false;
-	}
+	/*
+	 * Sheepdog's recovery code assumes every node have the same epoch
+	 * history. But we don't check epoch history of joining node because:
+	 * 1. inconsist epoch history only happens in the network partition case
+	 *    for the corosync driver, but corosync driver will panic for such
+	 *    case to prevent epoch inconsistency.
+	 * 2. checking epoch history with joining node is too expensive and is
+	 *    unneeded for zookeeper driver.
+	 *
+	 * That said, we don't check epoch history at all.
+	 */
 
 	return true;
 }
 
 main_fn void sd_accept_handler(const struct sd_node *joined,
-			       const struct sd_node *members, size_t nr_members,
+			       const struct rb_root *nroot, size_t nr_nodes,
 			       const void *opaque)
 {
-	int i;
 	const struct cluster_info *cinfo = opaque;
+	struct sd_node *n;
 
-	if (!cluster_join_check(cinfo)) {
+	if (node_is_local(joined) && !cluster_join_check(cinfo)) {
 		sd_err("failed to join Sheepdog");
 		exit(1);
 	}
 
-	sys->cinfo = *cinfo;
+	cluster_info_copy(&sys->cinfo, cinfo);
 
 	sd_debug("join %s", node_to_str(joined));
-	for (i = 0; i < nr_members; i++)
-		sd_debug("[%x] %s", i, node_to_str(members + i));
+	rb_for_each_entry(n, nroot, rb) {
+		sd_debug("%s", node_to_str(n));
+	}
 
 	if (sys->cinfo.status == SD_STATUS_SHUTDOWN)
 		return;
 
-	update_cluster_info(cinfo, joined, members, nr_members);
+	update_cluster_info(cinfo, joined, nroot, nr_nodes);
 
 	if (node_is_local(joined))
 		/* this output is used for testing */
@@ -898,15 +920,16 @@ main_fn void sd_accept_handler(const struct sd_node *joined,
 }
 
 main_fn void sd_leave_handler(const struct sd_node *left,
-			      const struct sd_node *members,
-			      size_t nr_members)
+			      const struct rb_root *nroot, size_t nr_nodes)
 {
 	struct vnode_info *old_vnode_info;
-	int i, ret;
+	struct sd_node *n;
+	int ret;
 
 	sd_debug("leave %s", node_to_str(left));
-	for (i = 0; i < nr_members; i++)
-		sd_debug("[%x] %s", i, node_to_str(members + i));
+	rb_for_each_entry(n, nroot, rb) {
+		sd_debug("%s", node_to_str(n));
+	}
 
 	if (sys->cinfo.status == SD_STATUS_SHUTDOWN)
 		return;
@@ -915,9 +938,12 @@ main_fn void sd_leave_handler(const struct sd_node *left,
 		/* Mark leave node as gateway only node */
 		sys->this_node.nr_vnodes = 0;
 
+	/*
+	 * Using main_thread_get() instead of get_vnode_info() is allowed
+	 * because of the same reason of update_cluster_info()
+	 */
 	old_vnode_info = main_thread_get(current_vnode_info);
-	main_thread_set(current_vnode_info,
-			  alloc_vnode_info(members, nr_members));
+	main_thread_set(current_vnode_info, alloc_vnode_info(nroot));
 	if (sys->cinfo.status == SD_STATUS_OK) {
 		ret = inc_and_log_epoch();
 		if (ret != 0)
@@ -933,19 +959,26 @@ main_fn void sd_leave_handler(const struct sd_node *left,
 
 static void update_node_size(struct sd_node *node)
 {
-	struct vnode_info *cur_vinfo = main_thread_get(current_vnode_info);
-	int idx = get_node_idx(cur_vinfo, node);
-	assert(idx != -1);
-	cur_vinfo->nodes[idx].space = node->space;
+	struct vnode_info *cur_vinfo = get_vnode_info();
+	struct sd_node *n = rb_search(&cur_vinfo->nroot, node, rb, node_cmp);
+
+	if (unlikely(!n))
+		panic("can't find %s", node_to_str(node));
+	n->space = node->space;
+
+	put_vnode_info(cur_vinfo);
 }
 
 static void kick_node_recover(void)
 {
+	/*
+	 * Using main_thread_get() instead of get_vnode_info() is allowed
+	 * because of the same reason of update_cluster_info()
+	 */
 	struct vnode_info *old = main_thread_get(current_vnode_info);
 	int ret;
 
-	main_thread_set(current_vnode_info,
-			alloc_vnode_info(old->nodes, old->nr_nodes));
+	main_thread_set(current_vnode_info, alloc_vnode_info(&old->nroot));
 	ret = inc_and_log_epoch();
 	if (ret != 0)
 		panic("cannot log current epoch %d", sys->cinfo.epoch);
@@ -1013,7 +1046,7 @@ int create_cluster(int port, int64_t zone, int nr_vnodes,
 	INIT_LIST_HEAD(&sys->local_req_queue);
 	INIT_LIST_HEAD(&sys->req_wait_queue);
 
-	ret = send_join_request(&sys->this_node);
+	ret = send_join_request();
 	if (ret != 0)
 		return -1;
 
