@@ -16,15 +16,16 @@
 /*
  * Object Cache ID
  *
- *  0 - 19 (20 bits): data object space
- *  20 - 27 (8 bits): object flag space
- *  28 - 31 (4 bits): object type indentifier space
+ *  0 - 31 (32 bits): data object space
+ *  32 - 51 (20 bits): reserved
+ *  52 - 59 (8 bits): object flag space
+ *  60 - 63 (4 bits): object type indentifier space
  */
-#define CACHE_VDI_SHIFT       31 /* if the entry is identified as VDI object */
-#define CACHE_CREATE_SHIFT    27 /* If the entry should be created at backend */
+#define CACHE_VDI_SHIFT       63 /* if the entry is identified as VDI object */
+#define CACHE_CREATE_SHIFT    59 /* If the entry should be created at backend */
 
-#define CACHE_VDI_BIT         (UINT32_C(1) << CACHE_VDI_SHIFT)
-#define CACHE_CREATE_BIT      (UINT32_C(1) << CACHE_CREATE_SHIFT)
+#define CACHE_VDI_BIT         (UINT64_C(1) << CACHE_VDI_SHIFT)
+#define CACHE_CREATE_BIT      (UINT64_C(1) << CACHE_CREATE_SHIFT)
 
 #define CACHE_INDEX_MASK      (CACHE_CREATE_BIT)
 
@@ -39,15 +40,15 @@ struct global_cache {
 };
 
 struct object_cache_entry {
-	uint32_t idx; /* Index of this entry */
+	uint64_t idx; /* Index of this entry */
 	refcnt_t refcnt; /* Reference count of this entry */
 	uint64_t bmap; /* Each bit represents one dirty block in object */
 	struct object_cache *oc; /* Object cache this entry belongs to */
 	struct rb_node node; /* For lru tree of object cache */
-	struct list_head dirty_list; /* For dirty list of object cache */
-	struct list_head lru_list; /* For lru list of object cache */
+	struct list_node dirty_list; /* For dirty list of object cache */
+	struct list_node lru_list; /* For lru list of object cache */
 
-	struct sd_lock lock; /* Entry lock */
+	struct sd_rw_lock lock; /* Entry lock */
 };
 
 struct object_cache {
@@ -60,9 +61,9 @@ struct object_cache {
 	struct list_head lru_head; /* Per VDI LRU list for reclaimer */
 	struct list_head dirty_head; /* Dirty objects linked to this list */
 	int push_efd; /* Used to synchronize between pusher and push threads */
-	uatomic_bool in_push; /* Whether if pusher is running */
+	struct sd_mutex push_mutex; /* mutex for pushing cache */
 
-	struct sd_lock lock; /* Cache lock */
+	struct sd_rw_lock lock; /* Cache lock */
 };
 
 struct push_work {
@@ -78,8 +79,8 @@ static int def_open_flags = O_RDWR;
 #define HASH_BITS	5
 #define HASH_SIZE	(1 << HASH_BITS)
 
-static struct sd_lock hashtable_lock[HASH_SIZE] = {
-	[0 ... HASH_SIZE - 1] = SD_LOCK_INITIALIZER
+static struct sd_rw_lock hashtable_lock[HASH_SIZE] = {
+	[0 ... HASH_SIZE - 1] = SD_RW_LOCK_INITIALIZER
 };
 
 static struct hlist_head cache_hashtable[HASH_SIZE];
@@ -97,20 +98,28 @@ static inline int hash(uint64_t vid)
 }
 
 /* We should always use this helper to get entry idx */
-static inline uint32_t entry_idx(const struct object_cache_entry *entry)
+static inline uint64_t entry_idx(const struct object_cache_entry *entry)
 {
 	return entry->idx & ~CACHE_INDEX_MASK;
 }
 
-static inline uint32_t object_cache_oid_to_idx(uint64_t oid)
+static int object_cache_cmp(const struct object_cache_entry *a,
+			    const struct object_cache_entry *b)
 {
-	uint32_t idx = data_oid_to_idx(oid);
+	return intcmp(entry_idx(a), entry_idx(b));
+}
+
+static inline uint64_t object_cache_oid_to_idx(uint64_t oid)
+{
+	uint64_t idx = data_oid_to_idx(oid);
 	if (is_vdi_obj(oid))
-		idx |= 1 << CACHE_VDI_SHIFT;
+		idx |= 1ULL << CACHE_VDI_SHIFT;
+	else if (is_vdi_btree_obj(oid))
+		idx |= VDI_BTREE_BIT;
 	return idx;
 }
 
-static inline bool idx_has_vdi_bit(uint32_t idx)
+static inline bool idx_has_vdi_bit(uint64_t idx)
 {
 	return !!(idx & CACHE_VDI_BIT);
 }
@@ -134,7 +143,7 @@ static uint64_t calc_object_bmap(uint64_t oid, size_t len, off_t offset)
 	nr = end - start;
 
 	while (nr--)
-		set_bit(start + nr, &bmap);
+		set_bit_64(start + nr, &bmap);
 
 	return bmap;
 }
@@ -178,7 +187,7 @@ static inline void write_lock_cache(struct object_cache *oc)
 
 static inline void unlock_cache(struct object_cache *oc)
 {
-	sd_unlock(&oc->lock);
+	sd_rw_unlock(&oc->lock);
 }
 
 static inline void read_lock_entry(struct object_cache_entry *entry)
@@ -193,54 +202,21 @@ static inline void write_lock_entry(struct object_cache_entry *entry)
 
 static inline void unlock_entry(struct object_cache_entry *entry)
 {
-	sd_unlock(&entry->lock);
+	sd_rw_unlock(&entry->lock);
 }
 
 static struct object_cache_entry *
 lru_tree_insert(struct rb_root *root, struct object_cache_entry *new)
 {
-	struct rb_node **p = &root->rb_node;
-	struct rb_node *parent = NULL;
-	struct object_cache_entry *entry;
-	uint32_t idx = entry_idx(new);
-
-	while (*p) {
-		parent = *p;
-		entry = rb_entry(parent, struct object_cache_entry, node);
-
-		if (idx < entry_idx(entry))
-			p = &(*p)->rb_left;
-		else if (idx > entry_idx(entry))
-			p = &(*p)->rb_right;
-		else {
-			/* already has this entry */
-			return entry;
-		}
-	}
-	rb_link_node(&new->node, parent, p);
-	rb_insert_color(&new->node, root);
-
-	return NULL; /* insert successfully */
+	return rb_insert(root, new, node, object_cache_cmp);
 }
 
 static struct object_cache_entry *lru_tree_search(struct rb_root *root,
-						  uint32_t idx)
+						  uint64_t idx)
 {
-	struct rb_node *n = root->rb_node;
-	struct object_cache_entry *t;
+	struct object_cache_entry key = { .idx = idx };
 
-	while (n) {
-		t = rb_entry(n, struct object_cache_entry, node);
-
-		if (idx < entry_idx(t))
-			n = n->rb_left;
-		else if (idx > entry_idx(t))
-			n = n->rb_right;
-		else
-			return t; /* found it */
-	}
-
-	return NULL;
+	return rb_search(root, &key, node, object_cache_cmp);
 }
 
 static void do_background_push(struct work *work)
@@ -248,11 +224,11 @@ static void do_background_push(struct work *work)
 	struct push_work *pw = container_of(work, struct push_work, work);
 	struct object_cache *oc = pw->oc;
 
-	if (!uatomic_set_true(&oc->in_push))
+	if (sd_mutex_trylock(&oc->push_mutex) == EBUSY)
 		return;
 
 	object_cache_push(oc);
-	uatomic_set_false(&oc->in_push);
+	sd_mutex_unlock(&oc->push_mutex);
 }
 
 static void background_push_done(struct work *work)
@@ -276,7 +252,7 @@ static void del_from_dirty_list(struct object_cache_entry *entry)
 {
 	struct object_cache *oc = entry->oc;
 
-	list_del_init(&entry->dirty_list);
+	list_del(&entry->dirty_list);
 	uatomic_dec(&oc->dirty_count);
 }
 
@@ -287,26 +263,24 @@ static void add_to_dirty_list(struct object_cache_entry *entry)
 	list_add_tail(&entry->dirty_list, &oc->dirty_head);
 	/* FIXME read sys->status atomically */
 	if (uatomic_add_return(&oc->dirty_count, 1) > MAX_DIRTY_OBJECT_COUNT
-	    && !uatomic_is_true(&oc->in_push)
 	    && sys->cinfo.status == SD_STATUS_OK)
 		kick_background_pusher(oc);
 }
 
-static inline void
-free_cache_entry(struct object_cache_entry *entry)
+static inline void free_cache_entry(struct object_cache_entry *entry)
 {
 	struct object_cache *oc = entry->oc;
 
 	rb_erase(&entry->node, &oc->lru_tree);
-	list_del_init(&entry->lru_list);
+	list_del(&entry->lru_list);
 	oc->total_count--;
-	if (!list_empty(&entry->dirty_list))
+	if (list_linked(&entry->dirty_list))
 		del_from_dirty_list(entry);
-	sd_destroy_lock(&entry->lock);
+	sd_destroy_rw_lock(&entry->lock);
 	free(entry);
 }
 
-static uint64_t idx_to_oid(uint32_t vid, uint32_t idx)
+static uint64_t idx_to_oid(uint32_t vid, uint64_t idx)
 {
 	if (idx_has_vdi_bit(idx))
 		return vid_to_vdi_oid(vid);
@@ -314,12 +288,12 @@ static uint64_t idx_to_oid(uint32_t vid, uint32_t idx)
 		return vid_to_data_oid(vid, idx);
 }
 
-static int remove_cache_object(struct object_cache *oc, uint32_t idx)
+static int remove_cache_object(struct object_cache *oc, uint64_t idx)
 {
 	int ret = SD_RES_SUCCESS;
 	char path[PATH_MAX];
 
-	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%08"PRIx32,
+	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%016"PRIx64,
 		 object_cache_dir, oc->vid, idx);
 	sd_debug("%"PRIx64, idx_to_oid(oc->vid, idx));
 	if (unlikely(unlink(path) < 0)) {
@@ -333,14 +307,14 @@ out:
 	return ret;
 }
 
-static int read_cache_object_noupdate(uint32_t vid, uint32_t idx, void *buf,
+static int read_cache_object_noupdate(uint32_t vid, uint64_t idx, void *buf,
 				      size_t count, off_t offset)
 {
 	size_t size;
 	int fd, flags = def_open_flags, ret = SD_RES_SUCCESS;
 	char p[PATH_MAX];
 
-	snprintf(p, sizeof(p), "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir,
+	snprintf(p, sizeof(p), "%s/%06"PRIx32"/%016"PRIx64, object_cache_dir,
 		 vid, idx);
 
 	if (sys->object_cache_directio && !idx_has_vdi_bit(idx)) {
@@ -370,14 +344,14 @@ out:
 	return ret;
 }
 
-static int write_cache_object_noupdate(uint32_t vid, uint32_t idx, void *buf,
+static int write_cache_object_noupdate(uint32_t vid, uint64_t idx, void *buf,
 				       size_t count, off_t offset)
 {
 	size_t size;
 	int fd, flags = def_open_flags, ret = SD_RES_SUCCESS;
 	char p[PATH_MAX];
 
-	snprintf(p, sizeof(p), "%s/%06"PRIx32"/%08"PRIx32, object_cache_dir,
+	snprintf(p, sizeof(p), "%s/%06"PRIx32"/%016"PRIx64, object_cache_dir,
 		 vid, idx);
 	if (sys->object_cache_directio && !idx_has_vdi_bit(idx)) {
 		assert(is_aligned_to_pagesize(buf));
@@ -409,7 +383,8 @@ out:
 static int read_cache_object(struct object_cache_entry *entry, void *buf,
 			     size_t count, off_t offset)
 {
-	uint32_t vid = entry->oc->vid, idx = entry_idx(entry);
+	uint32_t vid = entry->oc->vid;
+	uint64_t idx = entry_idx(entry);
 	struct object_cache *oc = entry->oc;
 	int ret;
 
@@ -427,7 +402,8 @@ static int write_cache_object(struct object_cache_entry *entry, void *buf,
 			      size_t count, off_t offset, bool create,
 			      bool writeback)
 {
-	uint32_t vid = entry->oc->vid, idx = entry_idx(entry);
+	uint32_t vid = entry->oc->vid;
+	uint64_t idx = entry_idx(entry);
 	uint64_t oid = idx_to_oid(vid, idx);
 	struct object_cache *oc = entry->oc;
 	struct sd_req hdr;
@@ -443,7 +419,7 @@ static int write_cache_object(struct object_cache_entry *entry, void *buf,
 	write_lock_cache(oc);
 	if (writeback) {
 		entry->bmap |= calc_object_bmap(oid, count, offset);
-		if (list_empty(&entry->dirty_list))
+		if (!list_linked(&entry->dirty_list))
 			add_to_dirty_list(entry);
 	}
 	list_move_tail(&entry->lru_list, &oc->lru_head);
@@ -474,7 +450,7 @@ out:
 	return ret;
 }
 
-static int push_cache_object(uint32_t vid, uint32_t idx, uint64_t bmap,
+static int push_cache_object(uint32_t vid, uint64_t idx, uint64_t bmap,
 			     bool create)
 {
 	struct sd_req hdr;
@@ -497,7 +473,7 @@ static int push_cache_object(uint32_t vid, uint32_t idx, uint64_t bmap,
 		 oid, bsize, bmap, first_bit, last_bit);
 	offset = first_bit * bsize;
 	data_length = min((last_bit - first_bit + 1) * bsize,
-			  get_objsize(oid) - offset);
+			  get_objsize(oid) - (size_t)offset);
 
 	buf = xvalloc(data_length);
 	ret = read_cache_object_noupdate(vid, idx, buf, data_length, offset);
@@ -538,12 +514,12 @@ out:
 #define HIGH_WATERMARK (sys->object_cache_size * 9 / 10)
 static void do_reclaim_object(struct object_cache *oc)
 {
-	struct object_cache_entry *entry, *t;
+	struct object_cache_entry *entry;
 	uint64_t oid;
 	uint32_t cap;
 
 	write_lock_cache(oc);
-	list_for_each_entry_safe(entry, t, &oc->lru_head, lru_list) {
+	list_for_each_entry(entry, &oc->lru_head, lru_list) {
 		oid = idx_to_oid(oc->vid, entry_idx(entry));
 		if (entry_in_use(entry)) {
 			sd_debug("%"PRIx64" is in use, skip...", oid);
@@ -598,12 +574,12 @@ static void do_reclaim(struct work *work)
 			do_reclaim_object(cache);
 			cap = uatomic_read(&gcache.capacity);
 			if (cap <= HIGH_WATERMARK) {
-				sd_unlock(&hashtable_lock[idx]);
+				sd_rw_unlock(&hashtable_lock[idx]);
 				sd_debug("complete, capacity %"PRIu32, cap);
 				return;
 			}
 		}
-		sd_unlock(&hashtable_lock[idx]);
+		sd_rw_unlock(&hashtable_lock[idx]);
 	}
 	sd_debug("finished");
 }
@@ -658,17 +634,19 @@ not_found:
 		INIT_LIST_HEAD(&cache->dirty_head);
 		INIT_LIST_HEAD(&cache->lru_head);
 
-		sd_init_lock(&cache->lock);
+		sd_init_rw_lock(&cache->lock);
 		hlist_add_head(&cache->hash, head);
+
+		sd_init_mutex(&cache->push_mutex);
 	} else {
 		cache = NULL;
 	}
 out:
-	sd_unlock(&hashtable_lock[h]);
+	sd_rw_unlock(&hashtable_lock[h]);
 	return cache;
 }
 
-void object_cache_try_to_reclaim(int delay)
+static void object_cache_try_to_reclaim(int delay)
 {
 	struct reclaim_work *rw;
 
@@ -690,21 +668,21 @@ void object_cache_try_to_reclaim(int delay)
 }
 
 static inline struct object_cache_entry *
-alloc_cache_entry(struct object_cache *oc, uint32_t idx)
+alloc_cache_entry(struct object_cache *oc, uint64_t idx)
 {
 	struct object_cache_entry *entry;
 
 	entry = xzalloc(sizeof(*entry));
 	entry->oc = oc;
 	entry->idx = idx;
-	sd_init_lock(&entry->lock);
-	INIT_LIST_HEAD(&entry->dirty_list);
-	INIT_LIST_HEAD(&entry->lru_list);
+	sd_init_rw_lock(&entry->lock);
+	INIT_LIST_NODE(&entry->dirty_list);
+	INIT_LIST_NODE(&entry->lru_list);
 
 	return entry;
 }
 
-static void add_to_lru_cache(struct object_cache *oc, uint32_t idx, bool create)
+static void add_to_lru_cache(struct object_cache *oc, uint64_t idx, bool create)
 {
 	struct object_cache_entry *entry = alloc_cache_entry(oc, idx);
 
@@ -740,13 +718,13 @@ static inline int lookup_path(char *path)
 	return ret;
 }
 
-static int object_cache_lookup(struct object_cache *oc, uint32_t idx,
+static int object_cache_lookup(struct object_cache *oc, uint64_t idx,
 			       bool create, bool writeback)
 {
 	int fd, ret, flags = def_open_flags;
 	char path[PATH_MAX];
 
-	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%08"PRIx32,
+	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%016"PRIx64,
 		 object_cache_dir, oc->vid, idx);
 	if (!create)
 		return lookup_path(path);
@@ -771,20 +749,19 @@ out:
 	return ret;
 }
 
-static int create_cache_object(struct object_cache *oc, uint32_t idx,
-			       void *buffer, size_t buf_size, off_t offset,
-			       size_t obj_size)
+static int create_cache_object(struct object_cache *oc, uint64_t idx,
+			       void *buffer, size_t buf_size)
 {
 	int flags = def_open_flags | O_CREAT | O_EXCL, fd;
 	int ret = SD_RES_OID_EXIST;
 	char path[PATH_MAX], tmp_path[PATH_MAX];
 
-	snprintf(tmp_path, sizeof(tmp_path), "%s/%06"PRIx32"/%08"PRIx32".tmp",
+	snprintf(tmp_path, sizeof(tmp_path), "%s/%06"PRIx32"/%016"PRIx64".tmp",
 		object_cache_dir, oc->vid, idx);
 	fd = open(tmp_path, flags, sd_def_fmode);
 	if (fd < 0) {
 		if (likely(errno == EEXIST)) {
-			sd_debug("%08"PRIx32" already created", idx);
+			sd_debug("%016"PRIx64" already created", idx);
 			goto out;
 		}
 		sd_debug("%m");
@@ -792,24 +769,14 @@ static int create_cache_object(struct object_cache *oc, uint32_t idx,
 		goto out;
 	}
 
-	/* We need to extend it if the buffer is trimmed */
-	if (offset != 0 || buf_size != obj_size) {
-		ret = prealloc(fd, obj_size);
-		if (unlikely(ret < 0)) {
-			ret = SD_RES_EIO;
-			sd_err("%m");
-			goto out_close;
-		}
-	}
-
-	ret = xpwrite(fd, buffer, buf_size, offset);
+	ret = xwrite(fd, buffer, buf_size);
 	if (unlikely(ret != buf_size)) {
 		ret = SD_RES_EIO;
-		sd_err("failed, vid %"PRIx32", idx %"PRIx32, oc->vid, idx);
+		sd_err("failed, vid %"PRIx32", idx %"PRIx64, oc->vid, idx);
 		goto out_close;
 	}
 	/* This is intended to take care of partial write due to crash */
-	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%08"PRIx32,
+	snprintf(path, sizeof(path), "%s/%06"PRIx32"/%016"PRIx64,
 		 object_cache_dir, oc->vid, idx);
 	ret = link(tmp_path, path);
 	if (unlikely(ret < 0)) {
@@ -823,7 +790,7 @@ static int create_cache_object(struct object_cache *oc, uint32_t idx,
 		goto out_close;
 	}
 	ret = SD_RES_SUCCESS;
-	sd_debug("%08"PRIx32" size %zu", idx, obj_size);
+	sd_debug("%016"PRIx64" size %zu", idx, buf_size);
 out_close:
 	close(fd);
 	unlink(tmp_path);
@@ -832,10 +799,9 @@ out:
 }
 
 /* Fetch the object, cache it in the clean state */
-static int object_cache_pull(struct object_cache *oc, uint32_t idx)
+static int object_cache_pull(struct object_cache *oc, uint64_t idx)
 {
 	struct sd_req hdr;
-	struct sd_rsp *rsp = (struct sd_rsp *)&hdr;
 	int ret = SD_RES_NO_MEM;
 	uint64_t oid = idx_to_oid(oc->vid, idx);
 	uint32_t data_length = get_objsize(oid);
@@ -851,8 +817,7 @@ static int object_cache_pull(struct object_cache *oc, uint32_t idx)
 		goto err;
 
 	sd_debug("oid %"PRIx64" pulled successfully", oid);
-	ret = create_cache_object(oc, idx, buf, rsp->data_length,
-				  rsp->obj.offset, data_length);
+	ret = create_cache_object(oc, idx, buf, data_length);
 	/*
 	 * We try to delay reclaim objects to avoid object ping-pong
 	 * because the pulled object is clean and likely to be reclaimed
@@ -927,7 +892,7 @@ static void push_object_done(struct work *work)
  */
 static int object_cache_push(struct object_cache *oc)
 {
-	struct object_cache_entry *entry, *t;
+	struct object_cache_entry *entry;
 
 	write_lock_cache(oc);
 	if (list_empty(&oc->dirty_head)) {
@@ -936,7 +901,7 @@ static int object_cache_push(struct object_cache *oc)
 	}
 
 	uatomic_set(&oc->push_count, uatomic_read(&oc->dirty_count));
-	list_for_each_entry_safe(entry, t, &oc->dirty_head, dirty_list) {
+	list_for_each_entry(entry, &oc->dirty_head, dirty_list) {
 		struct push_work *pw;
 
 		get_cache_entry(entry);
@@ -958,7 +923,7 @@ static int object_cache_push(struct object_cache *oc)
 bool object_is_cached(uint64_t oid)
 {
 	uint32_t vid = oid_to_vid(oid);
-	uint32_t idx = object_cache_oid_to_idx(oid);
+	uint64_t idx = object_cache_oid_to_idx(oid);
 	struct object_cache *cache;
 
 	cache = find_object_cache(vid, false);
@@ -972,7 +937,7 @@ void object_cache_delete(uint32_t vid)
 {
 	struct object_cache *cache;
 	int h = hash(vid);
-	struct object_cache_entry *entry, *t;
+	struct object_cache_entry *entry;
 	char path[PATH_MAX];
 
 	cache = find_object_cache(vid, false);
@@ -982,15 +947,15 @@ void object_cache_delete(uint32_t vid)
 	/* Firstly we free memeory */
 	sd_write_lock(&hashtable_lock[h]);
 	hlist_del(&cache->hash);
-	sd_unlock(&hashtable_lock[h]);
+	sd_rw_unlock(&hashtable_lock[h]);
 
 	write_lock_cache(cache);
-	list_for_each_entry_safe(entry, t, &cache->lru_head, lru_list) {
+	list_for_each_entry(entry, &cache->lru_head, lru_list) {
 		free_cache_entry(entry);
 		uatomic_sub(&gcache.capacity, CACHE_OBJECT_SIZE);
 	}
 	unlock_cache(cache);
-	sd_destroy_lock(&cache->lock);
+	sd_destroy_rw_lock(&cache->lock);
 	close(cache->push_efd);
 	free(cache);
 
@@ -1000,7 +965,7 @@ void object_cache_delete(uint32_t vid)
 }
 
 static struct object_cache_entry *
-get_cache_entry_from(struct object_cache *cache, uint32_t idx)
+get_cache_entry_from(struct object_cache *cache, uint64_t idx)
 {
 	struct object_cache_entry *entry;
 
@@ -1020,7 +985,7 @@ get_cache_entry_from(struct object_cache *cache, uint32_t idx)
 static struct object_cache_entry *oid_to_entry(uint64_t oid)
 {
 	uint32_t vid = oid_to_vid(oid);
-	uint32_t idx = object_cache_oid_to_idx(oid);
+	uint64_t idx = object_cache_oid_to_idx(oid);
 	struct object_cache *cache;
 	struct object_cache_entry *entry;
 
@@ -1038,7 +1003,7 @@ static int object_cache_flush_and_delete(struct object_cache *oc)
 	DIR *dir;
 	struct dirent *d;
 	uint32_t vid = oc->vid;
-	uint32_t idx;
+	uint64_t idx;
 	uint64_t all = UINT64_MAX;
 	int ret = 0;
 	char p[PATH_MAX];
@@ -1062,7 +1027,7 @@ static int object_cache_flush_and_delete(struct object_cache *oc)
 			continue;
 		}
 
-		idx = strtoul(d->d_name, NULL, 16);
+		idx = strtoull(d->d_name, NULL, 16);
 		if (idx == ULLONG_MAX)
 			continue;
 		if (push_cache_object(vid, idx, all, true) !=
@@ -1103,7 +1068,7 @@ bool bypass_object_cache(const struct request *req)
 			return true;
 		} else  {
 			/* For read requet, we can read cache if any */
-			uint32_t idx = object_cache_oid_to_idx(oid);
+			uint64_t idx = object_cache_oid_to_idx(oid);
 
 			if (object_cache_lookup(cache, idx, false, false) == 0)
 				return false;
@@ -1120,13 +1085,13 @@ int object_cache_handle_request(struct request *req)
 	struct sd_req *hdr = &req->rq;
 	uint64_t oid = req->rq.obj.oid;
 	uint32_t vid = oid_to_vid(oid);
-	uint32_t idx = object_cache_oid_to_idx(oid);
+	uint64_t idx = object_cache_oid_to_idx(oid);
 	struct object_cache *cache;
 	struct object_cache_entry *entry;
 	int ret;
 	bool create = false;
 
-	sd_debug("%08" PRIx32 ", len %" PRIu32 ", off %" PRIu64, idx,
+	sd_debug("%08" PRIx64 ", len %" PRIu32 ", off %" PRIu32, idx,
 		 hdr->data_length, hdr->obj.offset);
 
 	cache = find_object_cache(vid, true);
@@ -1226,11 +1191,10 @@ int object_cache_flush_vdi(uint32_t vid)
 	 * that dirty bits produced while it is waiting are guaranteed
 	 * to be pushed back
 	 */
-	while (!uatomic_set_true(&cache->in_push))
-		usleep(100000);
-
+	sd_mutex_lock(&cache->push_mutex);
 	ret = object_cache_push(cache);
-	uatomic_set_false(&cache->in_push);
+	sd_mutex_unlock(&cache->push_mutex);
+
 	return ret;
 }
 
@@ -1251,7 +1215,7 @@ static int load_cache_object(struct object_cache *cache)
 {
 	DIR *dir;
 	struct dirent *d;
-	uint32_t idx;
+	uint64_t idx;
 	char path[PATH_MAX];
 	int ret = 0;
 
@@ -1275,7 +1239,7 @@ static int load_cache_object(struct object_cache *cache)
 			continue;
 		}
 
-		idx = strtoul(d->d_name, NULL, 16);
+		idx = strtoull(d->d_name, NULL, 16);
 		if (idx == ULLONG_MAX)
 			continue;
 
@@ -1298,7 +1262,7 @@ static int load_cache(void)
 {
 	DIR *dir;
 	struct dirent *d;
-	uint32_t vid;
+	unsigned long vid;
 	char path[PATH_MAX];
 	int ret = 0;
 
@@ -1314,7 +1278,7 @@ static int load_cache(void)
 		if (!strncmp(d->d_name, ".", 1))
 			continue;
 		vid = strtoul(d->d_name, NULL, 16);
-		if (vid == ULLONG_MAX)
+		if (vid == ULONG_MAX)
 			continue;
 
 		load_cache_object(find_object_cache(vid, true));
@@ -1391,12 +1355,12 @@ err:
 void object_cache_format(void)
 {
 	struct object_cache *cache;
-	struct hlist_node *node, *t;
+	struct hlist_node *node;
 	int i;
 
 	for (i = 0; i < HASH_SIZE; i++) {
 		struct hlist_head *head = cache_hashtable + i;
-		hlist_for_each_entry_safe(cache, node, t, head, hash) {
+		hlist_for_each_entry(cache, node, head, hash) {
 			object_cache_delete(cache->vid);
 		}
 	}
@@ -1407,8 +1371,9 @@ int object_cache_get_info(struct object_cache_info *info)
 {
 	int j = 0;
 
-	info->used = gcache.capacity * 1024 * 1024;
-	info->size = sys->object_cache_size * 1024 * 1024;
+	memset(info, 0, sizeof(*info));
+	info->used = (uint64_t)gcache.capacity * 1024 * 1024;
+	info->size = (uint64_t)sys->object_cache_size * 1024 * 1024;
 
 	for (int i = 0; i < HASH_SIZE; i++) {
 		struct hlist_head *head = cache_hashtable + i;
@@ -1424,9 +1389,10 @@ int object_cache_get_info(struct object_cache_info *info)
 			j++;
 			unlock_cache(cache);
 		}
-		sd_unlock(&hashtable_lock[i]);
+		sd_rw_unlock(&hashtable_lock[i]);
 	}
 	info->count = j;
+	info->directio = sys->object_cache_directio;
 
 	return sizeof(*info);
 }
