@@ -26,6 +26,12 @@ struct sd_op_template {
 	/* process request even when cluster is not working */
 	bool force;
 
+	/*
+	 * Indicates administrative operation to trace.
+	 * If true is set, rx_main and tx_main log operations at info level.
+	 */
+	bool is_admin_op;
+
 	/*
 	 * process_work() will be called in a worker thread, and process_main()
 	 * will be called in the main thread.
@@ -46,6 +52,12 @@ struct sd_op_template {
 	int (*process_main)(const struct sd_req *req, struct sd_rsp *rsp, void *data);
 };
 
+/*
+ * The last gathered epoch is the epoch at which all the nodes complete the
+ * recovery and purge the stale objects.
+ */
+uint32_t last_gathered_epoch = 1;
+
 static int stat_sheep(uint64_t *store_size, uint64_t *store_free,
 		      uint32_t epoch)
 {
@@ -67,20 +79,38 @@ static int cluster_new_vdi(struct request *req)
 	struct sd_rsp *rsp = &req->rp;
 	uint32_t vid;
 	int ret;
+	struct timeval tv;
+
+	gettimeofday(&tv, NULL);
+
 	struct vdi_iocb iocb = {
 		.name = req->data,
 		.data_len = hdr->data_length,
 		.size = hdr->vdi.vdi_size,
 		.base_vid = hdr->vdi.base_vdi_id,
 		.create_snapshot = !!hdr->vdi.snapid,
-		.nr_copies = hdr->vdi.copies ? hdr->vdi.copies :
-				sys->cinfo.nr_copies,
+		.copy_policy = hdr->vdi.copy_policy,
+		.store_policy = hdr->vdi.store_policy,
+		.nr_copies = hdr->vdi.copies,
+		.time = (uint64_t) tv.tv_sec << 32 | tv.tv_usec * 1000,
 	};
 
+	/* Client doesn't specify redundancy scheme (copy = 0) */
+	if (!hdr->vdi.copies) {
+		iocb.nr_copies = sys->cinfo.nr_copies;
+		iocb.copy_policy = sys->cinfo.copy_policy;
+	}
+
+	if (iocb.copy_policy)
+		iocb.nr_copies = ec_policy_to_dp(iocb.copy_policy, NULL, NULL);
+
 	if (hdr->data_length != SD_MAX_VDI_LEN)
 		return SD_RES_INVALID_PARMS;
 
-	ret = vdi_create(&iocb, &vid);
+	if (iocb.create_snapshot)
+		ret = vdi_snapshot(&iocb, &vid);
+	else
+		ret = vdi_create(&iocb, &vid);
 
 	rsp->vdi.vdi_id = vid;
 	rsp->vdi.copies = iocb.nr_copies;
@@ -246,7 +276,8 @@ static int cluster_make_fs(const struct sd_req *req, struct sd_rsp *rsp,
 		return ret;
 
 	sys->cinfo.nr_copies = req->cluster.copies;
-	sys->cinfo.flags = req->flags;
+	sys->cinfo.copy_policy = req->cluster.copy_policy;
+	sys->cinfo.flags = req->cluster.flags;
 	if (!sys->cinfo.nr_copies)
 		sys->cinfo.nr_copies = SD_DEFAULT_COPIES;
 	sys->cinfo.ctime = req->cluster.ctime;
@@ -273,6 +304,17 @@ static int cluster_shutdown(const struct sd_req *req, struct sd_rsp *rsp,
 			    void *data)
 {
 	sys->cinfo.status = SD_STATUS_SHUTDOWN;
+	if (!node_in_recovery()) {
+		unregister_listening_fds();
+
+		if (set_cluster_shutdown(true) != SD_RES_SUCCESS)
+			/*
+			 * It's okay we failed to set 'shutdown', just start
+			 * recovery after restart blindly.
+			 */
+			sd_err("failed to set cluster as shutdown");
+	}
+
 	return SD_RES_SUCCESS;
 }
 
@@ -312,8 +354,7 @@ static int cluster_get_vdi_attr(struct request *req)
 	 * the current VDI id can change if we take a snapshot,
 	 * so we use the hash value of the VDI name as the VDI id
 	 */
-	vid = fnv_64a_buf(vattr->name, strlen(vattr->name), FNV1A_64_INIT);
-	vid &= SD_NR_VDIS - 1;
+	vid = sd_hash_vdi(vattr->name);
 	ret = get_vdi_attr(req->data, hdr->data_length,
 			   vid, &attrid, info.create_time,
 			   !!(hdr->flags & SD_FLAG_CMD_CREAT),
@@ -415,8 +456,18 @@ static int local_stat_cluster(struct request *req)
 
 		elog = (struct epoch_log *)req->data + i;
 		memset(elog, 0, sizeof(*elog));
+
+		/* some filed only need to store in first elog */
+		if (i == 0) {
+			elog->ctime = sys->cinfo.ctime;
+			elog->disable_recovery = sys->cinfo.disable_recovery;
+			elog->nr_copies = sys->cinfo.nr_copies;
+			elog->copy_policy = sys->cinfo.copy_policy;
+			strncpy(elog->drv_name, (char *)sys->cinfo.store,
+				STORE_LEN);
+		}
+
 		elog->epoch = epoch;
-		elog->ctime = sys->cinfo.ctime;
 		nr_nodes = epoch_log_read_with_timestamp(epoch, elog->nodes,
 							 sizeof(elog->nodes),
 							 (time_t *)&elog->time);
@@ -429,7 +480,6 @@ static int local_stat_cluster(struct request *req)
 		assert(nr_nodes <= SD_MAX_NODES);
 		elog->nr_nodes = nr_nodes;
 
-		elog->disable_recovery = sys->cinfo.disable_recovery;
 
 		rsp->data_length += sizeof(*elog);
 		epoch--;
@@ -499,15 +549,14 @@ static int cluster_force_recover_work(struct request *req)
 	}
 
 	if (req->rq.data_length <
-	    sizeof(*old_vnode_info->nodes) * old_vnode_info->nr_nodes) {
+	    sizeof(struct sd_node) * old_vnode_info->nr_nodes) {
 		sd_err("too small buffer size, %d", req->rq.data_length);
 		return SD_RES_INVALID_PARMS;
 	}
 
 	req->rp.epoch = epoch;
-	req->rp.data_length = sizeof(*old_vnode_info->nodes) *
-		old_vnode_info->nr_nodes;
-	memcpy(req->data, old_vnode_info->nodes, req->rp.data_length);
+	req->rp.data_length = sizeof(struct sd_node) * old_vnode_info->nr_nodes;
+	nodes_to_buffer(&old_vnode_info->nroot, req->data);
 
 	put_vnode_info(old_vnode_info);
 
@@ -522,6 +571,7 @@ static int cluster_force_recover_main(const struct sd_req *req,
 	int ret = SD_RES_SUCCESS;
 	struct sd_node *nodes = data;
 	size_t nr_nodes = rsp->data_length / sizeof(*nodes);
+	struct rb_root nroot = RB_ROOT;
 
 	if (rsp->epoch != sys->cinfo.epoch) {
 		sd_err("epoch was incremented while cluster_force_recover");
@@ -540,8 +590,11 @@ static int cluster_force_recover_main(const struct sd_req *req,
 
 	sys->cinfo.status = SD_STATUS_OK;
 
+	for (int i = 0; i < nr_nodes; i++)
+		rb_insert(&nroot, &nodes[i], rb, node_cmp);
+
 	vnode_info = get_vnode_info();
-	old_vnode_info = alloc_vnode_info(nodes, nr_nodes);
+	old_vnode_info = alloc_vnode_info(&nroot);
 	start_recovery(vnode_info, old_vnode_info, true);
 	put_vnode_info(vnode_info);
 	put_vnode_info(old_vnode_info);
@@ -576,12 +629,13 @@ static int cluster_notify_vdi_add(const struct sd_req *req, struct sd_rsp *rsp,
 		/* make the previous working vdi a snapshot */
 		add_vdi_state(req->vdi_state.old_vid,
 			      get_vdi_copy_number(req->vdi_state.old_vid),
-			      true);
+			      true, req->vdi_state.copy_policy);
 
 	if (req->vdi_state.set_bitmap)
 		atomic_set_bit(req->vdi_state.new_vid, sys->vdi_inuse);
 
-	add_vdi_state(req->vdi_state.new_vid, req->vdi_state.copies, false);
+	add_vdi_state(req->vdi_state.new_vid, req->vdi_state.copies, false,
+		      req->vdi_state.copy_policy);
 
 	return SD_RES_SUCCESS;
 }
@@ -641,13 +695,16 @@ static int cluster_recovery_completion(const struct sd_req *req,
 
 	if (vnode_info->nr_nodes == nr_recovereds) {
 		for (i = 0; i < nr_recovereds; ++i) {
-			if (!node_eq(vnode_info->nodes + i, recovereds + i))
+			if (!rb_search(&vnode_info->nroot, &recovereds[i],
+				       rb, node_cmp))
 				break;
 		}
 		if (i == nr_recovereds) {
-			sd_debug("all nodes are recovered, epoch %d", epoch);
+			sd_notice("all nodes are recovered, epoch %d", epoch);
+			last_gathered_epoch = epoch;
 			/* sd_store can be NULL if this node is a gateway */
-			if (sd_store && sd_store->cleanup)
+			if (vnode_info->nr_zones >= ec_max_data_strip &&
+			    sd_store && sd_store->cleanup)
 				sd_store->cleanup();
 		}
 	}
@@ -744,6 +801,28 @@ static int local_get_cache_info(struct request *request)
 	return SD_RES_SUCCESS;
 }
 
+static int local_cache_purge(struct request *req)
+{
+	const struct sd_req *hdr = &req->rq;
+	uint32_t vid = oid_to_vid(req->rq.obj.oid);
+
+	if (hdr->flags == SD_FLAG_CMD_WRITE) {
+		object_cache_delete(vid);
+		goto out;
+	}
+	object_cache_format();
+out:
+	return SD_RES_SUCCESS;
+}
+
+static int local_sd_stat(const struct sd_req *req, struct sd_rsp *rsp,
+			 void *data)
+{
+	memcpy(data, &sys->stat, sizeof(struct sd_stat));
+	rsp->data_length = sizeof(struct sd_stat);
+	return SD_RES_SUCCESS;
+}
+
 /* Return SD_RES_INVALID_PARMS to ask client not to send flush req again */
 static int local_flush_vdi(struct request *req)
 {
@@ -760,21 +839,34 @@ static int local_flush_vdi(struct request *req)
 static int local_discard_obj(struct request *req)
 {
 	uint64_t oid = req->rq.obj.oid;
-	uint32_t vid = oid_to_vid(oid), zero = 0;
-	int ret, idx = data_oid_to_idx(oid);
+	uint32_t vid = oid_to_vid(oid), tmp_vid;
+	int ret = SD_RES_SUCCESS, idx = data_oid_to_idx(oid);
+	struct sd_inode *inode = xmalloc(sizeof(struct sd_inode));
 
 	sd_debug("%"PRIx64, oid);
-	ret = write_object(vid_to_vdi_oid(vid), (char *)&zero, sizeof(zero),
-			   SD_INODE_HEADER_SIZE + sizeof(vid) * idx, false);
+	ret = sd_read_object(vid_to_vdi_oid(vid), (char *)inode,
+			     sizeof(struct sd_inode), 0);
 	if (ret != SD_RES_SUCCESS)
-		return ret;
-	if (remove_object(oid) != SD_RES_SUCCESS)
-		sd_err("failed to remove %"PRIx64, oid);
+		goto out;
+
+	tmp_vid = INODE_GET_VID(inode, idx);
+	/* if vid in idx is not exist, we don't need to remove it */
+	if (tmp_vid) {
+		INODE_SET_VID(inode, idx, 0);
+		ret = sd_inode_write_vid(sheep_bnode_writer, inode, idx, vid,
+					 0, 0, false, false);
+		if (ret != SD_RES_SUCCESS)
+			goto out;
+		if (sd_remove_object(oid) != SD_RES_SUCCESS)
+			sd_err("failed to remove %"PRIx64, oid);
+	}
 	/*
-	 * Return success even if remove_object fails because we have updated
+	 * Return success even if sd_remove_object fails because we have updated
 	 * inode successfully.
 	 */
-	return SD_RES_SUCCESS;
+out:
+	free(inode);
+	return ret;
 }
 
 static int local_flush_and_del(struct request *req)
@@ -823,47 +915,19 @@ static int local_kill_node(const struct sd_req *req, struct sd_rsp *rsp,
 			   void *data)
 {
 	sys->cinfo.status = SD_STATUS_KILLED;
+	unregister_listening_fds();
 
 	return SD_RES_SUCCESS;
 }
 
-static int read_copy_from_replica(struct request *req, uint32_t epoch,
-				  uint64_t oid, char *buf)
-{
-	struct request read_req = { };
-	struct sd_req *hdr = &read_req.rq;
-	struct sd_rsp *rsp = &read_req.rp;
-	int ret;
-
-	/* Create a fake gateway read request */
-	sd_init_req(hdr, SD_OP_READ_OBJ);
-	hdr->data_length = SD_DATA_OBJ_SIZE;
-	hdr->epoch = epoch;
-
-	hdr->obj.oid = oid;
-	hdr->obj.offset = 0;
-	hdr->obj.copies = get_req_copy_number(req);
-
-	read_req.data = buf;
-	read_req.op = get_sd_op(hdr->opcode);
-	read_req.vinfo = req->vinfo;
-
-	ret = gateway_read_obj(&read_req);
-
-	if (ret == SD_RES_SUCCESS)
-		untrim_zero_blocks(buf, rsp->obj.offset, rsp->data_length,
-				   SD_DATA_OBJ_SIZE);
-
-	return ret;
-}
-
-int peer_remove_obj(struct request *req)
+static int peer_remove_obj(struct request *req)
 {
 	uint64_t oid = req->rq.obj.oid;
+	uint8_t ec_index = req->rq.obj.ec_index;
 
 	objlist_cache_remove(oid);
 
-	return sd_store->remove_object(oid);
+	return sd_store->remove_object(oid, ec_index);
 }
 
 int peer_read_obj(struct request *req)
@@ -882,34 +946,18 @@ int peer_read_obj(struct request *req)
 	iocb.buf = req->data;
 	iocb.length = hdr->data_length;
 	iocb.offset = hdr->obj.offset;
+	iocb.ec_index = hdr->obj.ec_index;
+	iocb.copy_policy = hdr->obj.copy_policy;
 	ret = sd_store->read(hdr->obj.oid, &iocb);
 	if (ret != SD_RES_SUCCESS)
 		goto out;
 
 	rsp->data_length = hdr->data_length;
-	rsp->obj.offset = 0;
-	trim_zero_blocks(req->data, &rsp->obj.offset, &rsp->data_length);
-
-	if (hdr->obj.copies)
-		rsp->obj.copies = hdr->obj.copies;
-	else
-		rsp->obj.copies = get_obj_copy_number(hdr->obj.oid,
-						      req->vinfo->nr_zones);
 out:
 	return ret;
 }
 
-static int do_create_and_write_obj(struct siocb *iocb, struct sd_req *hdr,
-				   uint32_t epoch, void *data)
-{
-	iocb->buf = data;
-	iocb->length = hdr->data_length;
-	iocb->offset = hdr->obj.offset;
-
-	return sd_store->create_and_write(hdr->obj.oid, iocb);
-}
-
-int peer_write_obj(struct request *req)
+static int peer_write_obj(struct request *req)
 {
 	struct sd_req *hdr = &req->rq;
 	struct siocb iocb = { };
@@ -919,53 +967,54 @@ int peer_write_obj(struct request *req)
 	iocb.buf = req->data;
 	iocb.length = hdr->data_length;
 	iocb.offset = hdr->obj.offset;
+	iocb.ec_index = hdr->obj.ec_index;
+	iocb.copy_policy = hdr->obj.copy_policy;
 
 	return sd_store->write(oid, &iocb);
 }
 
-int peer_create_and_write_obj(struct request *req)
+static int peer_create_and_write_obj(struct request *req)
 {
 	struct sd_req *hdr = &req->rq;
-	struct sd_req cow_hdr;
-	uint32_t epoch = hdr->epoch;
-	uint64_t oid = hdr->obj.oid;
-	char *buf = NULL;
-	struct siocb iocb;
-	int ret = SD_RES_SUCCESS;
+	struct siocb iocb = { };
 
-	memset(&iocb, 0, sizeof(iocb));
-	iocb.epoch = epoch;
-	iocb.length = get_objsize(oid);
-	if (hdr->flags & SD_FLAG_CMD_COW) {
-		sd_debug("%" PRIx64 ", %" PRIx64, oid, hdr->obj.cow_oid);
+	iocb.epoch = hdr->epoch;
+	iocb.buf = req->data;
+	iocb.length = hdr->data_length;
+	iocb.ec_index = hdr->obj.ec_index;
+	iocb.copy_policy = hdr->obj.copy_policy;
+	iocb.offset = hdr->obj.offset;
 
-		buf = xvalloc(SD_DATA_OBJ_SIZE);
-		if (hdr->data_length != SD_DATA_OBJ_SIZE) {
-			ret = read_copy_from_replica(req, hdr->epoch,
-						     hdr->obj.cow_oid, buf);
-			if (ret != SD_RES_SUCCESS) {
-				sd_err("failed to read cow object");
-				goto out;
-			}
-		}
+	return sd_store->create_and_write(hdr->obj.oid, &iocb);
+}
 
-		memcpy(buf + hdr->obj.offset, req->data, hdr->data_length);
-		memcpy(&cow_hdr, hdr, sizeof(cow_hdr));
-		cow_hdr.data_length = SD_DATA_OBJ_SIZE;
-		cow_hdr.obj.offset = 0;
-		trim_zero_blocks(buf, &cow_hdr.obj.offset,
-				 &cow_hdr.data_length);
+static int local_get_loglevel(struct request *req)
+{
+	int32_t current_level;
 
-		ret = do_create_and_write_obj(&iocb, &cow_hdr, epoch, buf);
-	} else
-		ret = do_create_and_write_obj(&iocb, hdr, epoch, req->data);
+	current_level = get_loglevel();
+	memcpy(req->data, &current_level, sizeof(current_level));
+	req->rp.data_length = sizeof(current_level);
+
+	sd_info("returning log level: %u", current_level);
+
+	return SD_RES_SUCCESS;
+}
+
+static int local_set_loglevel(struct request *req)
+{
+	int32_t new_level = 0;
+
+	memcpy(&new_level, req->data, sizeof(int32_t));
+	if (!(LOG_EMERG <= new_level && new_level <= LOG_DEBUG)) {
+		sd_err("invalid log level: %d", new_level);
+		return SD_RES_INVALID_PARMS;
+	}
+
+	set_loglevel(new_level);
+
+	return SD_RES_SUCCESS;
 
-	if (SD_RES_SUCCESS == ret)
-		objlist_cache_insert(oid);
-out:
-	if (buf)
-		free(buf);
-	return ret;
 }
 
 static struct sd_op_template sd_ops[] = {
@@ -974,6 +1023,7 @@ static struct sd_op_template sd_ops[] = {
 	[SD_OP_NEW_VDI] = {
 		.name = "NEW_VDI",
 		.type = SD_OP_TYPE_CLUSTER,
+		.is_admin_op = true,
 		.process_work = cluster_new_vdi,
 		.process_main = post_cluster_new_vdi,
 	},
@@ -981,6 +1031,7 @@ static struct sd_op_template sd_ops[] = {
 	[SD_OP_DEL_VDI] = {
 		.name = "DEL_VDI",
 		.type = SD_OP_TYPE_CLUSTER,
+		.is_admin_op = true,
 		.process_work = cluster_del_vdi,
 		.process_main = post_cluster_del_vdi,
 	},
@@ -989,6 +1040,7 @@ static struct sd_op_template sd_ops[] = {
 		.name = "MAKE_FS",
 		.type = SD_OP_TYPE_CLUSTER,
 		.force = true,
+		.is_admin_op = true,
 		.process_main = cluster_make_fs,
 	},
 
@@ -996,6 +1048,7 @@ static struct sd_op_template sd_ops[] = {
 		.name = "SHUTDOWN",
 		.type = SD_OP_TYPE_CLUSTER,
 		.force = true,
+		.is_admin_op = true,
 		.process_main = cluster_shutdown,
 	},
 
@@ -1009,6 +1062,7 @@ static struct sd_op_template sd_ops[] = {
 		.name = "FORCE_RECOVER",
 		.type = SD_OP_TYPE_CLUSTER,
 		.force = true,
+		.is_admin_op = true,
 		.process_work = cluster_force_recover_work,
 		.process_main = cluster_force_recover_main,
 	},
@@ -1062,18 +1116,21 @@ static struct sd_op_template sd_ops[] = {
 	[SD_OP_REWEIGHT] = {
 		.name = "REWEIGHT",
 		.type = SD_OP_TYPE_CLUSTER,
+		.is_admin_op = true,
 		.process_main = cluster_reweight,
 	},
 
 	[SD_OP_ENABLE_RECOVER] = {
 		.name = "ENABLE_RECOVER",
 		.type = SD_OP_TYPE_CLUSTER,
+		.is_admin_op = true,
 		.process_main = cluster_enable_recover,
 	},
 
 	[SD_OP_DISABLE_RECOVER] = {
 		.name = "DISABLE_RECOVER",
 		.type = SD_OP_TYPE_CLUSTER,
+		.is_admin_op = true,
 		.process_main = cluster_disable_recover,
 	},
 
@@ -1193,6 +1250,7 @@ static struct sd_op_template sd_ops[] = {
 		.name = "KILL_NODE",
 		.type = SD_OP_TYPE_LOCAL,
 		.force = true,
+		.is_admin_op = true,
 		.process_main = local_kill_node,
 	},
 
@@ -1205,12 +1263,14 @@ static struct sd_op_template sd_ops[] = {
 	[SD_OP_MD_PLUG] = {
 		.name = "MD_PLUG_DISKS",
 		.type = SD_OP_TYPE_LOCAL,
+		.is_admin_op = true,
 		.process_main = local_md_plug,
 	},
 
 	[SD_OP_MD_UNPLUG] = {
 		.name = "MD_UNPLUG_DISKS",
 		.type = SD_OP_TYPE_LOCAL,
+		.is_admin_op = true,
 		.process_main = local_md_unplug,
 	},
 
@@ -1226,6 +1286,32 @@ static struct sd_op_template sd_ops[] = {
 		.process_work = local_get_cache_info,
 	},
 
+	[SD_OP_CACHE_PURGE] = {
+		.name = "CACHE_PURGE",
+		.type = SD_OP_TYPE_LOCAL,
+		.process_work = local_cache_purge,
+	},
+
+	[SD_OP_STAT] = {
+		.name = "STAT",
+		.type = SD_OP_TYPE_LOCAL,
+		.process_main = local_sd_stat,
+	},
+
+	[SD_OP_GET_LOGLEVEL] = {
+		.name = "GET_LOGLEVEL",
+		.type = SD_OP_TYPE_LOCAL,
+		.force = true,
+		.process_work = local_get_loglevel,
+	},
+
+	[SD_OP_SET_LOGLEVEL] = {
+		.name = "SET_LOGLEVEL",
+		.type = SD_OP_TYPE_LOCAL,
+		.force = true,
+		.process_work = local_set_loglevel,
+	},
+
 	/* gateway I/O operations */
 	[SD_OP_CREATE_AND_WRITE_OBJ] = {
 		.name = "CREATE_AND_WRITE_OBJ",
@@ -1287,42 +1373,50 @@ const struct sd_op_template *get_sd_op(uint8_t opcode)
 
 const char *op_name(const struct sd_op_template *op)
 {
+	if (op == NULL)
+		return "(invalid opcode)";
+
 	return op->name;
 }
 
 bool is_cluster_op(const struct sd_op_template *op)
 {
-	return op->type == SD_OP_TYPE_CLUSTER;
+	return op != NULL && op->type == SD_OP_TYPE_CLUSTER;
 }
 
 bool is_local_op(const struct sd_op_template *op)
 {
-	return op->type == SD_OP_TYPE_LOCAL;
+	return op != NULL && op->type == SD_OP_TYPE_LOCAL;
 }
 
 bool is_peer_op(const struct sd_op_template *op)
 {
-	return op->type == SD_OP_TYPE_PEER;
+	return op != NULL && op->type == SD_OP_TYPE_PEER;
 }
 
 bool is_gateway_op(const struct sd_op_template *op)
 {
-	return op->type == SD_OP_TYPE_GATEWAY;
+	return op != NULL && op->type == SD_OP_TYPE_GATEWAY;
 }
 
 bool is_force_op(const struct sd_op_template *op)
 {
-	return !!op->force;
+	return op != NULL && op->force;
+}
+
+bool is_logging_op(const struct sd_op_template *op)
+{
+	return op != NULL && op->is_admin_op;
 }
 
 bool has_process_work(const struct sd_op_template *op)
 {
-	return !!op->process_work;
+	return op != NULL && !!op->process_work;
 }
 
 bool has_process_main(const struct sd_op_template *op)
 {
-	return !!op->process_main;
+	return op != NULL && !!op->process_main;
 }
 
 void do_process_work(struct work *work)
@@ -1350,11 +1444,6 @@ int do_process_main(const struct sd_op_template *op, const struct sd_req *req,
 	return op->process_main(req, rsp, data);
 }
 
-int sheep_do_op_work(const struct sd_op_template *op, struct request *req)
-{
-	return op->process_work(req);
-}
-
 static int map_table[] = {
 	[SD_OP_CREATE_AND_WRITE_OBJ] = SD_OP_CREATE_AND_WRITE_PEER,
 	[SD_OP_READ_OBJ] = SD_OP_READ_PEER,
