@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2009-2011 Nippon Telegraph and Telephone Corporation.
+ * Copyright (C) 2012-2013 Taobao Inc.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version
@@ -82,8 +83,8 @@ static size_t list_buffer_size = DEFAULT_LIST_BUFFER_SIZE;
 
 static int obj_cmp(const uint64_t *oid1, const uint64_t *oid2)
 {
-	const uint64_t hval1 = fnv_64a_buf(oid1, sizeof(*oid1), FNV1A_64_INIT);
-	const uint64_t hval2 = fnv_64a_buf(oid2, sizeof(*oid2), FNV1A_64_INIT);
+	const uint64_t hval1 = sd_hash_oid(*oid1);
+	const uint64_t hval2 = sd_hash_oid(*oid2);
 
 	return intcmp(hval1, hval2);
 }
@@ -93,9 +94,138 @@ static inline bool node_is_gateway_only(void)
 	return sys->this_node.nr_vnodes == 0;
 }
 
-/* recover object from vnode */
+static struct vnode_info *rollback_vnode_info(uint32_t *epoch,
+					      struct vnode_info *cur)
+{
+	struct vnode_info *vinfo;
+rollback:
+	*epoch -= 1;
+	if (*epoch < last_gathered_epoch)
+		return NULL;
+
+	vinfo = get_vnode_info_epoch(*epoch, cur);
+	if (!vinfo) {
+		/* We rollback in case we don't get a valid epoch */
+		sd_alert("cannot get epoch %d", *epoch);
+		sd_alert("clients may see old data");
+		goto rollback;
+	}
+	return vinfo;
+}
+
+/*
+ * A node that does not match any node in current node list means the node has
+ * left the cluster, then it's an invalid node.
+ */
+static bool invalid_node(const struct sd_node *n, struct vnode_info *info)
+{
+
+	if (rb_search(&info->nroot, n, rb, node_cmp))
+		return false;
+	return true;
+}
+
+static int search_erasure_object(uint64_t oid, uint8_t idx,
+				 struct rb_root *nroot,
+				 struct recovery_work *rw,
+				 uint32_t tgt_epoch,
+				 void *buf)
+{
+	struct sd_req hdr;
+	unsigned rlen = get_store_objsize(oid);
+	struct sd_node *n;
+	uint32_t epoch = rw->epoch;
+
+	rb_for_each_entry(n, nroot, rb) {
+		if (invalid_node(n, rw->cur_vinfo))
+			continue;
+		sd_init_req(&hdr, SD_OP_READ_PEER);
+		hdr.epoch = epoch;
+		hdr.flags = SD_FLAG_CMD_RECOVERY;
+		hdr.data_length = rlen;
+		hdr.obj.oid = oid;
+		hdr.obj.tgt_epoch = tgt_epoch;
+		hdr.obj.ec_index = idx;
+
+		sd_debug("%"PRIx64" epoch %"PRIu32" tgt %"PRIu32" idx %d, %s",
+			 oid, epoch, tgt_epoch, idx, node_to_str(n));
+		if (sheep_exec_req(&n->nid, &hdr, buf) == SD_RES_SUCCESS)
+			return SD_RES_SUCCESS;
+	}
+	return SD_RES_NO_OBJ;
+}
+
+static void *read_erasure_object(uint64_t oid, uint8_t idx,
+				 struct recovery_obj_work *row)
+{
+	struct sd_req hdr;
+	unsigned rlen = get_store_objsize(oid);
+	void *buf = xvalloc(rlen);
+	struct recovery_work *rw = &row->base;
+	struct vnode_info *old = grab_vnode_info(rw->old_vinfo), *new_old;
+	uint32_t epoch = rw->epoch, tgt_epoch = rw->tgt_epoch;
+	const struct sd_node *node;
+	uint8_t policy = get_vdi_copy_policy(oid_to_vid(oid));
+	int edp = ec_policy_to_dp(policy, NULL, NULL);
+	int ret;
+again:
+	if (unlikely(old->nr_zones < edp)) {
+		if (search_erasure_object(oid, idx, &old->nroot, rw,
+					  tgt_epoch, buf)
+		    == SD_RES_SUCCESS)
+			goto done;
+		else
+			goto rollback;
+	}
+	node = oid_to_node(oid, &old->vroot, idx);
+	sd_debug("%"PRIx64" epoch %"PRIu32" tgt %"PRIu32" idx %d, %s",
+		 oid, epoch, tgt_epoch, idx, node_to_str(node));
+	if (invalid_node(node, rw->cur_vinfo))
+		goto rollback;
+	sd_init_req(&hdr, SD_OP_READ_PEER);
+	hdr.epoch = epoch;
+	hdr.flags = SD_FLAG_CMD_RECOVERY;
+	hdr.data_length = rlen;
+	hdr.obj.oid = oid;
+	hdr.obj.tgt_epoch = tgt_epoch;
+	hdr.obj.ec_index = idx;
+
+	ret = sheep_exec_req(&node->nid, &hdr, buf);
+	switch (ret) {
+	case SD_RES_SUCCESS:
+		goto done;
+	case SD_RES_OLD_NODE_VER:
+		free(buf);
+		buf = NULL;
+		row->stop = true;
+		break;
+	default:
+rollback:
+		new_old = rollback_vnode_info(&tgt_epoch, rw->cur_vinfo);
+		if (!new_old) {
+			sd_err("can not read %"PRIx64" idx %d", oid, idx);
+			free(buf);
+			buf = NULL;
+			goto done;
+		}
+		put_vnode_info(old);
+		old = new_old;
+		goto again;
+	}
+done:
+	put_vnode_info(old);
+	return buf;
+}
+
+/*
+ * Read object from targeted node and store it in the local node.
+ *
+ * tgt_epoch: the specific epoch that the object has stayed
+ * idx: erasure index. For non-erasure object, pass 0.
+ */
 static int recover_object_from(struct recovery_obj_work *row,
-			       const struct sd_node *node, uint32_t tgt_epoch)
+			       const struct sd_node *node,
+			       uint32_t tgt_epoch)
 {
 	uint64_t oid = row->oid;
 	uint32_t local_epoch = row->local_epoch;
@@ -133,7 +263,7 @@ static int recover_object_from(struct recovery_obj_work *row,
 		}
 	}
 
-	rlen = get_objsize(oid);
+	rlen = get_store_objsize(oid);
 	buf = xvalloc(rlen);
 
 	/* recover from remote replica */
@@ -157,18 +287,6 @@ static int recover_object_from(struct recovery_obj_work *row,
 	return ret;
 }
 
-/*
- * A node that does not match any node in current node list means the node has
- * left the cluster, then it's an invalid node.
- */
-static bool invalid_node(const struct sd_node *n, struct vnode_info *info)
-{
-
-	if (xbsearch(n, info->nodes, info->nr_nodes, node_cmp))
-		return false;
-	return true;
-}
-
 static int recover_object_from_replica(struct recovery_obj_work *row,
 				       struct vnode_info *old,
 				       uint32_t tgt_epoch)
@@ -184,7 +302,7 @@ static int recover_object_from_replica(struct recovery_obj_work *row,
 	for (int i = 0; i < nr_copies; i++) {
 		const struct sd_vnode *vnode;
 
-		vnode = oid_to_vnode(old->vnodes, old->nr_vnodes, oid, i);
+		vnode = oid_to_vnode(oid, &old->vroot, i);
 
 		if (vnode_is_local(vnode)) {
 			start = i;
@@ -197,8 +315,7 @@ static int recover_object_from_replica(struct recovery_obj_work *row,
 		const struct sd_node *node;
 		int idx = (i + start) % nr_copies;
 
-		node = oid_to_node(old->vnodes, old->nr_vnodes, oid, idx,
-				   old->nodes);
+		node = oid_to_node(oid, &old->vroot, idx);
 
 		if (invalid_node(node, row->base.cur_vinfo))
 			continue;
@@ -208,7 +325,6 @@ static int recover_object_from_replica(struct recovery_obj_work *row,
 		case SD_RES_SUCCESS:
 			sd_debug("recovered oid %"PRIx64" from %d to epoch %d",
 				 oid, tgt_epoch, epoch);
-			objlist_cache_insert(oid);
 			return ret;
 		case SD_RES_OLD_NODE_VER:
 			/* move to the next epoch recovery */
@@ -238,7 +354,7 @@ static int recover_object_from_replica(struct recovery_obj_work *row,
  * the routine will try to recovery it from the nodes it has stayed,
  * at least, *theoretically* on consistent hash ring.
  */
-static int do_recover_object(struct recovery_obj_work *row)
+static int recover_replication_object(struct recovery_obj_work *row)
 {
 	struct recovery_work *rw = &row->base;
 	struct vnode_info *old;
@@ -268,31 +384,138 @@ again:
 		/* fall through */
 	default:
 		/* No luck, roll back to an older configuration and try again */
-rollback:
-		tgt_epoch--;
-		if (tgt_epoch < 1) {
+		new_old = rollback_vnode_info(&tgt_epoch, rw->cur_vinfo);
+		if (!new_old) {
 			sd_err("can not recover oid %"PRIx64, oid);
 			ret = -1;
-			break;
-		}
-
-		new_old = get_vnode_info_epoch(tgt_epoch, rw->cur_vinfo);
-		if (!new_old) {
-			/* We rollback in case we don't get a valid epoch */
-			sd_alert("cannot get epoch %d", tgt_epoch);
-			sd_alert("clients may see old data");
-			goto rollback;
+			goto out;
 		}
 
 		put_vnode_info(old);
 		old = new_old;
 		goto again;
 	}
-
+out:
 	put_vnode_info(old);
 	return ret;
 }
 
+static void *rebuild_erasure_object(uint64_t oid, uint8_t idx,
+				    struct recovery_obj_work *row)
+{
+	int len = get_store_objsize(oid);
+	char *lost = xvalloc(len);
+	int i, j;
+	uint8_t policy = get_vdi_copy_policy(oid_to_vid(oid));
+	int ed = 0, edp;
+	edp = ec_policy_to_dp(policy, &ed, NULL);
+	struct fec *ctx = ec_init(ed, edp);
+	uint8_t *bufs[ed];
+	int idxs[ed];
+
+	for (i = 0; i < ed; i++)
+		bufs[i] = NULL;
+	for (i = 0; i < ed; i++)
+		idxs[i] = 0;
+
+	/* Prepare replica */
+	for (i = 0, j = 0; i < edp && j < ed; i++) {
+		if (i == idx)
+			continue;
+		bufs[j] = read_erasure_object(oid, i, row);
+		if (row->stop)
+			break;
+		if (!bufs[j])
+			continue;
+		idxs[j++] = i;
+	}
+	if (j != ed) {
+		free(lost);
+		lost = NULL;
+		goto out;
+	}
+
+	/* Rebuild the lost replica */
+	ec_decode_buffer(ctx, bufs, idxs, lost, idx);
+out:
+	ec_destroy(ctx);
+	for (i = 0; i < ed; i++)
+		free(bufs[i]);
+	return lost;
+}
+
+static uint8_t local_node_copy_index(struct vnode_info *vinfo, uint64_t oid)
+{
+	int idx;
+
+	if (!is_erasure_oid(oid))
+		return 0; /* no need to proceed */
+
+	for (idx = 0; idx < vinfo->nr_zones; idx++) {
+		const struct sd_node *n = oid_to_node(oid, &vinfo->vroot, idx);
+		if (node_is_local(n))
+			return idx;
+	}
+	panic("can't get valid index for %"PRIx64, oid);
+}
+
+/*
+ * Erasure object recovery algorithm
+ *
+ * 1. read the lost object from its track in epoch history vertically because
+ *    every copy that holds partial data of the object is unique
+ * 2. if not found in 1, then tries to rebuild it with RS algorithm
+ *    2.1 read enough other copies from their tracks in epoch history
+ *    2.2 rebuild the lost object from the content of copies read at 2.1
+ *
+ * The subtle case is number for available zones is less than total copy number
+ * or the requested index of lost object:
+ *    1 we need to make sure nr_zones >= total_copy_nr to avoid panic of
+ *      oid_to_node(s) helpers.
+ *    2 we have to search all the available zones when we can't get idx. Its
+ *      okay to do a mad search when number of available zones is small
+ */
+static int recover_erasure_object(struct recovery_obj_work *row)
+{
+	struct recovery_work *rw = &row->base;
+	struct vnode_info *cur = rw->cur_vinfo;
+	uint64_t oid = row->oid;
+	struct siocb iocb = { 0 };
+	void *buf = NULL;
+	uint8_t idx;
+	int ret = -1;
+
+	idx = local_node_copy_index(cur, oid);
+	buf = read_erasure_object(oid, idx, row);
+	if (!buf && !row->stop)
+		buf = rebuild_erasure_object(oid, idx, row);
+	if (!buf) {
+		if (!row->stop)
+			sd_err("failed to recover %"PRIx64" idx %d", oid, idx);
+		goto out;
+	}
+
+	iocb.epoch = rw->epoch;
+	iocb.length = get_store_objsize(oid);
+	iocb.offset = 0;
+	iocb.buf = buf;
+	iocb.ec_index = idx;
+	ret = sd_store->create_and_write(oid, &iocb);
+	free(buf);
+out:
+	return ret;
+}
+
+static int do_recover_object(struct recovery_obj_work *row)
+{
+	uint64_t oid = row->oid;
+
+	if (is_erasure_oid(oid))
+		return recover_erasure_object(row);
+	else
+		return recover_replication_object(row);
+}
+
 static void recover_object_work(struct work *work)
 {
 	struct recovery_work *rw = container_of(work, struct recovery_work,
@@ -301,25 +524,29 @@ static void recover_object_work(struct work *work)
 						     struct recovery_obj_work,
 						     base);
 	uint64_t oid = row->oid;
+	struct vnode_info *cur = rw->cur_vinfo;
 	int ret, epoch;
 
-	if (sd_store->exist(oid)) {
+	if (sd_store->exist(oid, local_node_copy_index(cur, oid))) {
 		sd_debug("the object is already recovered");
 		return;
 	}
 
 	/* find object in the stale directory */
-	for (epoch = sys_epoch() - 1; epoch > 0; epoch--) {
-		ret = sd_store->get_hash(oid, epoch, row->local_sha1);
-		if (ret == SD_RES_SUCCESS) {
-			sd_debug("replica found in local at epoch %d", epoch);
-			row->local_epoch = epoch;
-			break;
+	if (!is_erasure_oid(oid))
+		for (epoch = sys_epoch() - 1; epoch >= last_gathered_epoch;
+		     epoch--) {
+			ret = sd_store->get_hash(oid, epoch, row->local_sha1);
+			if (ret == SD_RES_SUCCESS) {
+				sd_debug("replica found in local at epoch %d",
+					 epoch);
+				row->local_epoch = epoch;
+				break;
+			}
 		}
-	}
 
 	ret = do_recover_object(row);
-	if (ret < 0)
+	if (ret != 0)
 		sd_err("failed to recover object %"PRIx64, oid);
 }
 
@@ -331,24 +558,11 @@ bool node_in_recovery(void)
 static inline void prepare_schedule_oid(uint64_t oid)
 {
 	struct recovery_info *rinfo = main_thread_get(current_rinfo);
-	uint64_t i;
 
-	for (i = 0; i < rinfo->nr_prio_oids; i++)
-		if (rinfo->prio_oids[i] == oid)
-			return;
-	/*
-	 * We need this check because oid might not be recovered.
-	 * Very much unlikely though, but it might happen indeed.
-	 */
-	for (i = 0; i < rinfo->done; i++)
-		if (rinfo->oids[i] == oid) {
-			sd_debug("%"PRIx64" not recovered, don't schedule it",
-				 oid);
-			return;
-		}
-	/* When recovery is not suspended, oid is currently being recovered */
-	if (!rinfo->suspended && rinfo->oids[rinfo->done] == oid)
+	if (xlfind(&oid, rinfo->prio_oids, rinfo->nr_prio_oids, oid_cmp)) {
+		sd_debug("%" PRIx64 " has been already in prio_oids", oid);
 		return;
+	}
 
 	rinfo->nr_prio_oids++;
 	rinfo->prio_oids = xrealloc(rinfo->prio_oids,
@@ -359,41 +573,65 @@ static inline void prepare_schedule_oid(uint64_t oid)
 	resume_suspended_recovery();
 }
 
-bool oid_in_recovery(uint64_t oid)
+main_fn bool oid_in_recovery(uint64_t oid)
 {
 	struct recovery_info *rinfo = main_thread_get(current_rinfo);
-	uint64_t i;
+	struct vnode_info *cur;
 
 	if (!node_in_recovery())
 		return false;
 
-	if (sd_store->exist(oid)) {
+	cur = rinfo->cur_vinfo;
+	if (sd_store->exist(oid, local_node_copy_index(cur, oid))) {
 		sd_debug("the object %" PRIx64 " is already recoverd", oid);
 		return false;
 	}
 
 	if (uatomic_read(&next_rinfo))
+		/*
+		 * The current recovery_info will be taken over by the next one
+		 * soon, so no need to call prepare_schedule_oid() now.
+		 */
 		return true;
 
-	/* If we are in preparation of object list, oid is not recovered yet */
-	if (rinfo->state == RW_PREPARE_LIST)
-		return true;
+	switch (rinfo->state) {
+	case RW_PREPARE_LIST:
+		/* oid is not recovered yet */
+		break;
+	case RW_RECOVER_OBJ:
+		if (xlfind(&oid, rinfo->oids, rinfo->done, oid_cmp)) {
+			sd_debug("%" PRIx64 " has been already recovered", oid);
+			return false;
+		}
 
-	/*
-	 * Check if oid is in the list that to be recovered later
-	 *
-	 * FIXME: do we need more efficient yet complex data structure?
-	 */
-	for (i = rinfo->done; i < rinfo->count; i++)
-		if (rinfo->oids[i] == oid)
+		if (rinfo->oids[rinfo->done] == oid) {
+			if (rinfo->suspended)
+				break;
+			/*
+			 * When recovery is not suspended,
+			 * rinfo->oids[rinfo->done] is currently being recovered
+			 * and no need to call prepare_schedule_oid().
+			 */
+			return true;
+		}
+
+		/*
+		 * Check if oid is in the list that to be recovered later
+		 *
+		 * FIXME: do we need more efficient yet complex data structure?
+		 */
+		if (xlfind(&oid, rinfo->oids + rinfo->done + 1,
+			   rinfo->count - (rinfo->done + 1), oid_cmp))
 			break;
 
-	/*
-	 * Newly created object after prepare_object_list() might not be
-	 * in the list
-	 */
-	if (i == rinfo->count) {
-		sd_err("%"PRIx64" is not in the recovery list", oid);
+		/*
+		 * Newly created object after prepare_object_list() might not be
+		 * in the list
+		 */
+		sd_debug("%"PRIx64" is not in the recovery list", oid);
+		return false;
+	case RW_NOTIFY_COMPLETION:
+		sd_debug("the object %" PRIx64 " is already recoverd", oid);
 		return false;
 	}
 
@@ -715,13 +953,8 @@ static void screen_object_list(struct recovery_list_work *rlw,
 			continue;
 
 		nr_objs = get_obj_copy_number(oids[i], rw->cur_vinfo->nr_zones);
-		if (!nr_objs) {
-			sd_err("ERROR: can not find copy number for object %"
-			       PRIx64, oids[i]);
-			continue;
-		}
-		oid_to_vnodes(rw->cur_vinfo->vnodes, rw->cur_vinfo->nr_vnodes,
-			      oids[i], nr_objs, vnodes);
+
+		oid_to_vnodes(oids[i], &rw->cur_vinfo->vroot, nr_objs, vnodes);
 		for (j = 0; j < nr_objs; j++) {
 			if (!vnode_is_local(vnodes[j]))
 				continue;
@@ -748,25 +981,28 @@ static void prepare_object_list(struct work *work)
 	struct recovery_list_work *rlw = container_of(rw,
 						      struct recovery_list_work,
 						      base);
-	struct sd_node *cur = rw->cur_vinfo->nodes;
-	int cur_nr = rw->cur_vinfo->nr_nodes;
-	int start = random() % cur_nr, i, end = cur_nr;
+	int nr_nodes = rw->cur_vinfo->nr_nodes;
+	int start = random() % nr_nodes, i, end = nr_nodes;
 	uint64_t *oids;
+	struct sd_node *nodes;
 
 	if (node_is_gateway_only())
 		return;
 
 	sd_debug("%u", rw->epoch);
 	wait_get_vdis_done();
+
+	nodes = xmalloc(sizeof(struct sd_node) * nr_nodes);
+	nodes_to_buffer(&rw->cur_vinfo->nroot, nodes);
 again:
 	/* We need to start at random node for better load balance */
 	for (i = start; i < end; i++) {
 		size_t nr_oids;
-		struct sd_node *node = cur + i;
+		struct sd_node *node = nodes + i;
 
 		if (uatomic_read(&next_rinfo)) {
 			sd_debug("go to the next recovery");
-			return;
+			goto out;
 		}
 
 		oids = fetch_object_list(node, rw->epoch, &nr_oids);
@@ -783,6 +1019,8 @@ again:
 	}
 
 	sd_debug("%"PRIu64, rlw->count);
+out:
+	free(nodes);
 }
 
 int start_recovery(struct vnode_info *cur_vinfo, struct vnode_info *old_vinfo,
