@@ -28,9 +28,7 @@ static bool is_access_local(struct request *req, uint64_t oid)
 	int i;
 
 	nr_copies = get_req_copy_number(req);
-	oid_to_vnodes(req->vinfo->vnodes, req->vinfo->nr_vnodes, oid,
-		      nr_copies, obj_vnodes);
-
+	oid_to_vnodes(oid, &req->vinfo->vroot, nr_copies, obj_vnodes);
 	for (i = 0; i < nr_copies; i++) {
 		if (vnode_is_local(obj_vnodes[i]))
 			return true;
@@ -180,13 +178,20 @@ static bool request_in_recovery(struct request *req)
 	    req->rq.opcode == SD_OP_CREATE_AND_WRITE_OBJ)
 		return false;
 
-	/*
-	 * Request from recovery should go down the Farm even if
-	 * oid_in_recovery() returns true because we should also try snap
-	 * cache of the Farm and return the error code back if not found.
-	 */
-	if (oid_in_recovery(req->local_oid) &&
-	    !(req->rq.flags & SD_FLAG_CMD_RECOVERY)) {
+	if (req->rq.flags & SD_FLAG_CMD_RECOVERY)
+		/*
+		 * Recovery requests must not be linked to wait queue to avoid a
+		 * dead lock.  Here is an example scenario.
+		 *  1. Node A sends a recovery request to node B.
+		 *  2. Node B links the request to the wait queue.
+		 *  3. Node B sends a recovery request to node A to recover the
+		 *     object.
+		 *  4. Node A links the request to the wait queue, and the
+		 *     object cannot be recovered on either A or B (dead lock).
+		 */
+		return false;
+
+	if (oid_in_recovery(req->local_oid)) {
 		sd_debug("%"PRIx64" wait on oid", req->local_oid);
 		sleep_on_wait_queue(req);
 		return true;
@@ -197,12 +202,12 @@ static bool request_in_recovery(struct request *req)
 /* Wakeup requests because of epoch mismatch */
 void wakeup_requests_on_epoch(void)
 {
-	struct request *req, *t;
+	struct request *req;
 	LIST_HEAD(pending_list);
 
 	list_splice_init(&sys->req_wait_queue, &pending_list);
 
-	list_for_each_entry_safe(req, t, &pending_list, request_list) {
+	list_for_each_entry(req, &pending_list, request_list) {
 		switch (req->rp.result) {
 		case SD_RES_OLD_NODE_VER:
 			/*
@@ -234,12 +239,12 @@ void wakeup_requests_on_epoch(void)
 /* Wakeup the requests on the oid that was previously being recoverred */
 void wakeup_requests_on_oid(uint64_t oid)
 {
-	struct request *req, *t;
+	struct request *req;
 	LIST_HEAD(pending_list);
 
 	list_splice_init(&sys->req_wait_queue, &pending_list);
 
-	list_for_each_entry_safe(req, t, &pending_list, request_list) {
+	list_for_each_entry(req, &pending_list, request_list) {
 		if (req->local_oid != oid)
 			continue;
 		sd_debug("retry %" PRIx64, req->local_oid);
@@ -250,12 +255,12 @@ void wakeup_requests_on_oid(uint64_t oid)
 
 void wakeup_all_requests(void)
 {
-	struct request *req, *n;
+	struct request *req;
 	LIST_HEAD(pending_list);
 
 	list_splice_init(&sys->req_wait_queue, &pending_list);
 
-	list_for_each_entry_safe(req, n, &pending_list, request_list) {
+	list_for_each_entry(req, &pending_list, request_list) {
 		sd_debug("%"PRIx64, req->rq.obj.oid);
 		del_requeue_request(req);
 	}
@@ -279,6 +284,22 @@ static void queue_peer_request(struct request *req)
 	queue_work(sys->io_wqueue, &req->work);
 }
 
+/*
+ * We make sure we write the exact number of copies to honor the promise of the
+ * redundancy for strict mode. This means that after writing of targeted data,
+ * they are redundant as promised and can withstand the random node failures.
+ *
+ * For example, with a 4:2 policy, we need at least write to 6 nodes with data
+ * strip and parity strips. For non-strict mode, we allow to write successfully
+ * only if the data are written fully with 4 nodes alive.
+ */
+static bool has_enough_zones(struct request *req)
+{
+	uint64_t oid = req->rq.obj.oid;
+
+	return req->vinfo->nr_zones >= get_vdi_copy_number(oid_to_vid(oid));
+}
+
 static void queue_gateway_request(struct request *req)
 {
 	struct sd_req *hdr = &req->rq;
@@ -303,9 +324,27 @@ static void queue_gateway_request(struct request *req)
 			return;
 
 queue_work:
+	if (RB_EMPTY_ROOT(&req->vinfo->vroot)) {
+		sd_err("there is no living nodes");
+		goto end_request;
+	}
+	if (sys->cinfo.flags & SD_CLUSTER_FLAG_STRICT &&
+	    (hdr->opcode == SD_OP_CREATE_AND_WRITE_OBJ ||
+	     hdr->opcode == SD_OP_WRITE_OBJ) &&
+	    !has_enough_zones(req)) {
+		sd_err("not enough zones available");
+		goto end_request;
+	}
+
 	req->work.fn = do_process_work;
 	req->work.done = gateway_op_done;
 	queue_work(sys->gateway_wqueue, &req->work);
+	return;
+
+end_request:
+	req->rp.result = SD_RES_HALT;
+	put_request(req);
+	return;
 }
 
 static void queue_local_request(struct request *req)
@@ -315,6 +354,74 @@ static void queue_local_request(struct request *req)
 	queue_work(sys->io_wqueue, &req->work);
 }
 
+static main_fn inline void stat_request_begin(struct request *req)
+{
+	struct sd_req *hdr = &req->rq;
+
+	req->stat = true;
+
+	if (is_peer_op(req->op)) {
+		sys->stat.r.peer_total_nr++;
+		sys->stat.r.peer_active_nr++;
+		if (hdr->flags & SD_FLAG_CMD_WRITE)
+			sys->stat.r.peer_total_rx += hdr->data_length;
+		else
+			sys->stat.r.peer_total_tx += hdr->data_length;
+
+		switch (hdr->opcode) {
+		case SD_OP_READ_PEER:
+			sys->stat.r.peer_total_read_nr++;
+			break;
+		case SD_OP_WRITE_PEER:
+		case SD_OP_CREATE_AND_WRITE_PEER:
+			sys->stat.r.peer_total_write_nr++;
+			break;
+		case SD_OP_REMOVE_PEER:
+			sys->stat.r.peer_total_remove_nr++;
+			break;
+		}
+	} else if (is_gateway_op(req->op)) {
+		sys->stat.r.gway_total_nr++;
+		sys->stat.r.gway_active_nr++;
+		if (hdr->flags & SD_FLAG_CMD_WRITE)
+			sys->stat.r.gway_total_rx += hdr->data_length;
+		else
+			sys->stat.r.gway_total_tx += hdr->data_length;
+
+		switch (hdr->opcode) {
+		case SD_OP_READ_OBJ:
+			sys->stat.r.gway_total_read_nr++;
+			break;
+		case SD_OP_WRITE_OBJ:
+		case SD_OP_CREATE_AND_WRITE_OBJ:
+			sys->stat.r.gway_total_write_nr++;
+			break;
+		case SD_OP_DISCARD_OBJ:
+			sys->stat.r.gway_total_remove_nr++;
+			break;
+		}
+	} else if (hdr->opcode == SD_OP_FLUSH_VDI) {
+		sys->stat.r.gway_total_nr++;
+		sys->stat.r.gway_active_nr++;
+		sys->stat.r.gway_total_flush_nr++;
+	}
+}
+
+static main_fn inline void stat_request_end(struct request *req)
+{
+	struct sd_req *hdr = &req->rq;
+
+	if (!req->stat)
+		return;
+
+	if (is_peer_op(req->op))
+		sys->stat.r.peer_active_nr--;
+	else if (is_gateway_op(req->op))
+		sys->stat.r.gway_active_nr--;
+	else if (hdr->opcode == SD_OP_FLUSH_VDI)
+		sys->stat.r.gway_active_nr--;
+}
+
 static void queue_request(struct request *req)
 {
 	struct sd_req *hdr = &req->rq;
@@ -367,7 +474,7 @@ static void queue_request(struct request *req)
 	}
 
 	req->vinfo = get_vnode_info();
-
+	stat_request_begin(req);
 	if (is_peer_op(req->op)) {
 		queue_peer_request(req);
 	} else if (is_gateway_op(req->op)) {
@@ -396,6 +503,7 @@ static void requeue_request(struct request *req)
 		put_vnode_info(req->vinfo);
 		req->vinfo = NULL;
 	}
+	stat_request_end(req);
 	queue_request(req);
 }
 
@@ -413,8 +521,6 @@ static struct request *alloc_local_request(void *data, int data_length)
 
 	req->local = true;
 
-	INIT_LIST_HEAD(&req->request_list);
-
 	refcount_set(&req->refcnt, 1);
 
 	return req;
@@ -426,6 +532,15 @@ static void free_local_request(struct request *req)
 	free(req);
 }
 
+static void submit_local_request(struct request *req)
+{
+	sd_mutex_lock(&sys->local_req_lock);
+	list_add_tail(&req->request_list, &sys->local_req_queue);
+	sd_mutex_unlock(&sys->local_req_lock);
+
+	eventfd_xwrite(sys->local_req_efd, 1);
+}
+
 /*
  * Exec the request locally and synchronously.
  *
@@ -441,16 +556,13 @@ worker_fn int exec_local_req(struct sd_req *rq, void *data)
 	req->rq = *rq;
 	req->local_req_efd = eventfd(0, 0);
 	if (req->local_req_efd < 0) {
+		sd_err("eventfd failed, %m");
 		/* Fake the result to ask for retry */
 		req->rp.result = SD_RES_NETWORK_ERROR;
 		goto out;
 	}
 
-	pthread_mutex_lock(&sys->local_req_lock);
-	list_add_tail(&req->request_list, &sys->local_req_queue);
-	pthread_mutex_unlock(&sys->local_req_lock);
-
-	eventfd_xwrite(sys->local_req_efd, 1);
+	submit_local_request(req);
 	eventfd_xread(req->local_req_efd);
 out:
 	/* fill rq with response header as exec_req does */
@@ -463,6 +575,79 @@ out:
 	return ret;
 }
 
+worker_fn struct request_iocb *local_req_init(void)
+{
+	struct request_iocb *iocb = xzalloc(sizeof(*iocb));
+
+	iocb->efd = eventfd(0, EFD_SEMAPHORE);
+	if (iocb->efd < 0) {
+		sd_err("eventfd failed, %m");
+		free(iocb);
+		return NULL;
+	}
+	iocb->result = SD_RES_SUCCESS;
+	return iocb;
+}
+
+worker_fn int local_req_wait(struct request_iocb *iocb)
+{
+	int ret;
+
+	for (uint32_t i = 0; i < iocb->count; i++)
+		eventfd_xread(iocb->efd);
+
+	ret = iocb->result;
+	close(iocb->efd);
+	free(iocb);
+	return ret;
+}
+
+struct areq_work {
+	struct sd_req rq;
+	void *data;
+	struct request_iocb *iocb;
+	int result;
+
+	struct work work;
+};
+
+static void local_req_async_work(struct work *work)
+{
+	struct areq_work *areq = container_of(work, struct areq_work, work);
+
+	areq->result = exec_local_req(&areq->rq, areq->data);
+}
+
+static void local_req_async_main(struct work *work)
+{
+	struct areq_work *areq = container_of(work, struct areq_work, work);
+
+	if (unlikely(areq->result != SD_RES_SUCCESS))
+		areq->iocb->result = areq->result;
+
+	eventfd_xwrite(areq->iocb->efd, 1);
+	free(areq);
+}
+
+worker_fn int exec_local_req_async(struct sd_req *rq, void *data,
+				   struct request_iocb *iocb)
+{
+	struct areq_work *areq;
+
+	areq = xzalloc(sizeof(*areq));
+	areq->rq = *rq;
+	areq->data = data;
+	areq->iocb = iocb;
+	areq->work.fn = local_req_async_work;
+	areq->work.done = local_req_async_main;
+
+	queue_work(sys->areq_wqueue, &areq->work);
+
+	iocb->count++;
+
+	return SD_RES_SUCCESS;
+}
+
 static struct request *alloc_request(struct client_info *ci, int data_length)
 {
 	struct request *req;
@@ -482,7 +667,6 @@ static struct request *alloc_request(struct client_info *ci, int data_length)
 		}
 	}
 
-	INIT_LIST_HEAD(&req->request_list);
 	refcount_set(&req->refcnt, 1);
 
 	uatomic_inc(&sys->nr_outstanding_reqs);
@@ -507,199 +691,156 @@ main_fn void put_request(struct request *req)
 	if (refcount_dec(&req->refcnt) > 0)
 		return;
 
+	stat_request_end(req);
+
 	if (req->local)
 		eventfd_xwrite(req->local_req_efd, 1);
 	else {
-		if (conn_tx_on(&ci->conn)) {
+		if (ci->conn.dead) {
 			clear_client_info(ci);
 			free_request(req);
 		} else {
-			list_add(&req->request_list, &ci->done_reqs);
+			list_add_tail(&req->request_list, &ci->done_reqs);
+
+			if (ci->tx_req == NULL)
+				/* There is no request being sent. */
+				conn_tx_on(&ci->conn);
 		}
 	}
 }
 
-static void init_rx_hdr(struct client_info *ci)
-{
-	ci->conn.c_rx_state = C_IO_HEADER;
-	ci->rx_req = NULL;
-	ci->conn.rx_length = sizeof(struct sd_req);
-	ci->conn.rx_buf = &ci->conn.rx_hdr;
-}
-
-static inline int begin_rx(struct client_info *ci)
+static void rx_work(struct work *work)
 {
+	struct client_info *ci = container_of(work, struct client_info,
+					      rx_work);
 	int ret;
-	uint64_t data_len;
 	struct connection *conn = &ci->conn;
-	struct sd_req *hdr = &conn->rx_hdr;
+	struct sd_req hdr;
 	struct request *req;
 
-	switch (conn->c_rx_state) {
-	case C_IO_HEADER:
-		ret = rx(conn, C_IO_DATA_INIT);
-		if (!ret || conn->c_rx_state != C_IO_DATA_INIT)
-			break;
-	case C_IO_DATA_INIT:
-		data_len = hdr->data_length;
-
-		req = alloc_request(ci, data_len);
-		if (!req) {
-			conn->c_rx_state = C_IO_CLOSED;
-			break;
-		}
-		ci->rx_req = req;
-
-		/* use le_to_cpu */
-		memcpy(&req->rq, hdr, sizeof(req->rq));
-
-		if (data_len && hdr->flags & SD_FLAG_CMD_WRITE) {
-			conn->c_rx_state = C_IO_DATA;
-			conn->rx_length = data_len;
-			conn->rx_buf = req->data;
-		} else {
-			conn->c_rx_state = C_IO_END;
-			break;
-		}
-	case C_IO_DATA:
-		ret = rx(conn, C_IO_END);
-		break;
-	default:
-		sd_err("bug: unknown state %d", conn->c_rx_state);
+	ret = do_read(conn->fd, &hdr, sizeof(hdr), NULL, 0, UINT32_MAX);
+	if (ret) {
+		sd_debug("failed to read a header");
+		conn->dead = true;
+		return;
 	}
 
-	if (is_conn_dead(conn)) {
-		clear_client_info(ci);
-		return -1;
+	req = alloc_request(ci, hdr.data_length);
+	if (!req) {
+		sd_err("failed to allocate request");
+		conn->dead = true;
+		return;
 	}
+	ci->rx_req = req;
 
-	/* Short read happens */
-	if (conn->c_rx_state != C_IO_END)
-		return -1;
+	/* use le_to_cpu */
+	memcpy(&req->rq, &hdr, sizeof(req->rq));
 
-	return 0;
+	if (hdr.data_length && hdr.flags & SD_FLAG_CMD_WRITE) {
+		ret = do_read(conn->fd, req->data, hdr.data_length, NULL, 0,
+			      UINT32_MAX);
+		if (ret) {
+			sd_err("failed to read data");
+			conn->dead = true;
+		}
+	}
 }
 
-static inline void finish_rx(struct client_info *ci)
+static void rx_main(struct work *work)
 {
-	struct request *req;
+	struct client_info *ci = container_of(work, struct client_info,
+					      rx_work);
+	struct request *req = ci->rx_req;
 
-	req = ci->rx_req;
-	init_rx_hdr(ci);
+	ci->rx_req = NULL;
 
-	sd_debug("%d, %s:%d", ci->conn.fd, ci->conn.ipstr, ci->conn.port);
+	refcount_dec(&ci->refcnt);
+
+	if (ci->conn.dead) {
+		if (req)
+			free_request(req);
+
+		clear_client_info(ci);
+		return;
+	}
+
+	conn_rx_on(&ci->conn);
+
+	if (is_logging_op(get_sd_op(req->rq.opcode))) {
+		sd_info("req=%p, fd=%d, client=%s:%d, op=%s, data=%s",
+			req,
+			ci->conn.fd,
+			ci->conn.ipstr, ci->conn.port,
+			op_name(get_sd_op(req->rq.opcode)),
+			data_to_str(req->data, req->rp.data_length));
+	} else {
+		sd_debug("%d, %s:%d",
+			 ci->conn.fd,
+			 ci->conn.ipstr,
+			 ci->conn.port);
+	}
 	queue_request(req);
 }
 
-static void do_client_rx(struct client_info *ci)
+static void tx_work(struct work *work)
 {
-	if (begin_rx(ci) < 0)
-		return;
-
-	finish_rx(ci);
-}
-
-static void init_tx_hdr(struct client_info *ci)
-{
-	struct sd_rsp *rsp = (struct sd_rsp *)&ci->conn.tx_hdr;
-	struct request *req;
-
-	assert(!list_empty(&ci->done_reqs));
-
-	memset(rsp, 0, sizeof(*rsp));
-
-	req = list_first_entry(&ci->done_reqs, struct request, request_list);
-	list_del(&req->request_list);
-
-	ci->tx_req = req;
-	ci->conn.tx_length = sizeof(*rsp);
-	ci->conn.c_tx_state = C_IO_HEADER;
-	ci->conn.tx_buf = rsp;
+	struct client_info *ci = container_of(work, struct client_info,
+					      tx_work);
+	int ret;
+	struct connection *conn = &ci->conn;
+	struct sd_rsp rsp;
+	struct request *req = ci->tx_req;
+	void *data = NULL;
 
 	/* use cpu_to_le */
-	memcpy(rsp, &req->rp, sizeof(*rsp));
+	memcpy(&rsp, &req->rp, sizeof(rsp));
 
-	rsp->epoch = sys->cinfo.epoch;
-	rsp->opcode = req->rq.opcode;
-	rsp->id = req->rq.id;
+	rsp.epoch = sys->cinfo.epoch;
+	rsp.opcode = req->rq.opcode;
+	rsp.id = req->rq.id;
+
+	if (rsp.data_length)
+		data = req->data;
+
+	ret = send_req(conn->fd, (struct sd_req *)&rsp, data, rsp.data_length,
+		       NULL, 0, UINT32_MAX);
+	if (ret != 0) {
+		sd_err("failed to send a request");
+		conn->dead = true;
+	}
 }
 
-static inline int begin_tx(struct client_info *ci)
+static void tx_main(struct work *work)
 {
-	int ret, opt;
-	struct sd_rsp *rsp = (struct sd_rsp *)&ci->conn.tx_hdr;
+	struct client_info *ci = container_of(work, struct client_info,
+					      tx_work);
 
-	/* If short send happens, we don't need init hdr */
-	if (!ci->tx_req)
-		init_tx_hdr(ci);
+	refcount_dec(&ci->refcnt);
 
-	opt = 1;
-	setsockopt(ci->conn.fd, SOL_TCP, TCP_CORK, &opt, sizeof(opt));
-
-	switch (ci->conn.c_tx_state) {
-	case C_IO_HEADER:
-		ret = tx(&ci->conn, C_IO_DATA_INIT);
-		if (!ret)
-			break;
-
-		if (rsp->data_length) {
-			ci->conn.tx_length = rsp->data_length;
-			ci->conn.tx_buf = ci->tx_req->data;
-			ci->conn.c_tx_state = C_IO_DATA;
-		} else {
-			ci->conn.c_tx_state = C_IO_END;
-			break;
-		}
-	case C_IO_DATA:
-		ret = tx(&ci->conn, C_IO_END);
-		if (!ret)
-			break;
-	default:
-		break;
+	if (is_logging_op(ci->tx_req->op)) {
+		sd_info("req=%p, fd=%d, client=%s:%d, op=%s, result=%02X",
+			ci->tx_req,
+			ci->conn.fd,
+			ci->conn.ipstr,
+			ci->conn.port,
+			op_name(ci->tx_req->op),
+			ci->tx_req->rp.result);
+	} else {
+		sd_debug("%d, %s:%d",
+			 ci->conn.fd,
+			 ci->conn.ipstr,
+			 ci->conn.port);
 	}
+	free_request(ci->tx_req);
+	ci->tx_req = NULL;
 
-	opt = 0;
-	setsockopt(ci->conn.fd, SOL_TCP, TCP_CORK, &opt, sizeof(opt));
-
-	if (is_conn_dead(&ci->conn)) {
+	if (ci->conn.dead) {
 		clear_client_info(ci);
-		return -1;
-	}
-	return 0;
-}
-
-/* Return 1 if short send happens or we have more data to send */
-static inline int finish_tx(struct client_info *ci)
-{
-	/* Finish sending one response */
-	if (ci->conn.c_tx_state == C_IO_END) {
-		sd_debug("connection from: %d, %s:%d", ci->conn.fd,
-			 ci->conn.ipstr, ci->conn.port);
-		free_request(ci->tx_req);
-		ci->tx_req = NULL;
-	}
-	if (ci->tx_req || !list_empty(&ci->done_reqs))
-		return 1;
-	return 0;
-}
-
-static void do_client_tx(struct client_info *ci)
-{
-	if (!ci->tx_req && list_empty(&ci->done_reqs)) {
-		if (conn_tx_off(&ci->conn))
-			clear_client_info(ci);
 		return;
 	}
 
-	if (begin_tx(ci) < 0)
-		return;
-
-	if (finish_tx(ci))
-		return;
-
-	/* Let's go sleep, and put_request() will wake me up */
-	if (conn_tx_off(&ci->conn))
-		clear_client_info(ci);
+	if (!list_empty(&ci->done_reqs))
+		conn_tx_on(&ci->conn);
 }
 
 static void destroy_client(struct client_info *ci)
@@ -711,21 +852,11 @@ static void destroy_client(struct client_info *ci)
 
 static void clear_client_info(struct client_info *ci)
 {
-	struct request *req, *t;
+	struct request *req;
 
 	sd_debug("connection seems to be dead");
 
-	if (ci->rx_req) {
-		free_request(ci->rx_req);
-		ci->rx_req = NULL;
-	}
-
-	if (ci->tx_req) {
-		free_request(ci->tx_req);
-		ci->tx_req = NULL;
-	}
-
-	list_for_each_entry_safe(req, t, &ci->done_reqs, request_list) {
+	list_for_each_entry(req, &ci->done_reqs, request_list) {
 		list_del(&req->request_list);
 		free_request(req);
 	}
@@ -751,8 +882,10 @@ static struct client_info *create_client(int fd, struct cluster_info *cluster)
 	if (!ci)
 		return NULL;
 
-	if (getpeername(fd, (struct sockaddr *)&from, &namesize))
+	if (getpeername(fd, (struct sockaddr *)&from, &namesize)) {
+		free(ci);
 		return NULL;
+	}
 
 	switch (from.ss_family) {
 	case AF_INET:
@@ -773,8 +906,6 @@ static struct client_info *create_client(int fd, struct cluster_info *cluster)
 
 	INIT_LIST_HEAD(&ci->done_reqs);
 
-	init_rx_hdr(ci);
-
 	return ci;
 }
 
@@ -782,17 +913,43 @@ static void client_handler(int fd, int events, void *data)
 {
 	struct client_info *ci = (struct client_info *)data;
 
-	sd_debug("%x, rx %d, tx %d", events, ci->conn.c_rx_state,
-		 ci->conn.c_tx_state);
+	sd_debug("%x, %d", events, ci->conn.dead);
 
-	if (events & (EPOLLERR | EPOLLHUP) || is_conn_dead(&ci->conn))
+	if (events & (EPOLLERR | EPOLLHUP) || ci->conn.dead)
 		return clear_client_info(ci);
 
-	if (events & EPOLLIN)
-		do_client_rx(ci);
+	if (events & EPOLLIN) {
+		if (conn_rx_off(&ci->conn) != 0)
+			return;
 
-	if (events & EPOLLOUT)
-		do_client_tx(ci);
+		/*
+		 * Increment refcnt so that the client_info isn't freed while
+		 * rx_work uses it.
+		 */
+		refcount_inc(&ci->refcnt);
+		ci->rx_work.fn = rx_work;
+		ci->rx_work.done = rx_main;
+		queue_work(sys->net_wqueue, &ci->rx_work);
+	}
+
+	if (events & EPOLLOUT) {
+		if (conn_tx_off(&ci->conn) != 0)
+			return;
+
+		assert(ci->tx_req == NULL);
+		ci->tx_req = list_first_entry(&ci->done_reqs, struct request,
+					      request_list);
+		list_del(&ci->tx_req->request_list);
+
+		/*
+		 * Increment refcnt so that the client_info isn't freed while
+		 * tx_work uses it.
+		 */
+		refcount_inc(&ci->refcnt);
+		ci->tx_work.fn = tx_work;
+		ci->tx_work.done = tx_main;
+		queue_work(sys->net_wqueue, &ci->tx_work);
+	}
 }
 
 static void listen_handler(int listen_fd, int events, void *data)
@@ -824,12 +981,6 @@ static void listen_handler(int listen_fd, int events, void *data)
 		}
 	}
 
-	ret = set_nonblocking(fd);
-	if (ret) {
-		close(fd);
-		return;
-	}
-
 	ci = create_client(fd, data);
 	if (!ci) {
 		close(fd);
@@ -845,12 +996,35 @@ static void listen_handler(int listen_fd, int events, void *data)
 	sd_debug("accepted a new connection: %d", fd);
 }
 
+static LIST_HEAD(listening_fd_list);
+
+struct listening_fd {
+	int fd;
+	struct list_node list;
+};
+
 static int create_listen_port_fn(int fd, void *data)
 {
+	struct listening_fd *new_fd;
+
+	new_fd = xzalloc(sizeof(*new_fd));
+	new_fd->fd = fd;
+	list_add_tail(&new_fd->list, &listening_fd_list);
+
 	return register_event(fd, listen_handler, data);
 }
 
-int create_listen_port(char *bindaddr, int port)
+void unregister_listening_fds(void)
+{
+	struct listening_fd *fd;
+
+	list_for_each_entry(fd, &listening_fd_list, list) {
+		sd_debug("unregistering fd: %d", fd->fd);
+		unregister_event(fd->fd);
+	}
+}
+
+int create_listen_port(const char *bindaddr, int port)
 {
 	static bool is_inet_socket = true;
 
@@ -872,7 +1046,7 @@ int init_unix_domain_socket(const char *dir)
 
 static void local_req_handler(int listen_fd, int events, void *data)
 {
-	struct request *req, *t;
+	struct request *req;
 	LIST_HEAD(pending_list);
 
 	if (events & EPOLLERR)
@@ -880,19 +1054,19 @@ static void local_req_handler(int listen_fd, int events, void *data)
 
 	eventfd_xread(listen_fd);
 
-	pthread_mutex_lock(&sys->local_req_lock);
+	sd_mutex_lock(&sys->local_req_lock);
 	list_splice_init(&sys->local_req_queue, &pending_list);
-	pthread_mutex_unlock(&sys->local_req_lock);
+	sd_mutex_unlock(&sys->local_req_lock);
 
-	list_for_each_entry_safe(req, t, &pending_list, request_list) {
+	list_for_each_entry(req, &pending_list, request_list) {
 		list_del(&req->request_list);
 		queue_request(req);
 	}
 }
 
-void local_req_init(void)
+void local_request_init(void)
 {
-	pthread_mutex_init(&sys->local_req_lock, NULL);
+	sd_init_mutex(&sys->local_req_lock);
 	sys->local_req_efd = eventfd(0, EFD_NONBLOCK);
 	if (sys->local_req_efd < 0)
 		panic("failed to init local req efd");
