@@ -42,6 +42,7 @@
 #include "sha1.h"
 #include "config.h"
 #include "sockfd_cache.h"
+#include "fec.h"
 
  /*
   * Functions that update global info must be called in the main
@@ -66,8 +67,10 @@ struct client_info {
 	struct connection conn;
 
 	struct request *rx_req;
+	struct work rx_work;
 
 	struct request *tx_req;
+	struct work tx_work;
 
 	struct list_head done_reqs;
 
@@ -81,6 +84,12 @@ enum REQUST_STATUS {
 	REQUEST_DROPPED
 };
 
+struct request_iocb {
+	uint32_t count;
+	int efd;
+	int result;
+};
+
 struct request {
 	struct sd_req rq;
 	struct sd_rsp rp;
@@ -91,8 +100,8 @@ struct request {
 	unsigned int data_length;
 
 	struct client_info *ci;
-	struct list_head request_list;
-	struct list_head pending_list;
+	struct list_node request_list;
+	struct list_node pending_list;
 
 	refcnt_t refcnt;
 	bool local;
@@ -104,6 +113,7 @@ struct request {
 
 	struct work work;
 	enum REQUST_STATUS status;
+	bool stat; /* true if this request is during stat */
 };
 
 struct system_info {
@@ -120,7 +130,7 @@ struct system_info {
 
 	int local_req_efd;
 
-	pthread_mutex_t local_req_lock;
+	struct sd_mutex local_req_lock;
 	struct list_head local_req_queue;
 	struct list_head req_wait_queue;
 	int nr_outstanding_reqs;
@@ -128,6 +138,7 @@ struct system_info {
 	bool gateway_only;
 	bool nosync;
 
+	struct work_queue *net_wqueue;
 	struct work_queue *gateway_wqueue;
 	struct work_queue *io_wqueue;
 	struct work_queue *deletion_wqueue;
@@ -137,6 +148,7 @@ struct system_info {
 	struct work_queue *oc_reclaim_wqueue;
 	struct work_queue *oc_push_wqueue;
 	struct work_queue *md_wqueue;
+	struct work_queue *areq_wqueue;
 #ifdef HAVE_HTTP
 	struct work_queue *http_wqueue;
 #endif
@@ -150,15 +162,19 @@ struct system_info {
 	bool backend_dio;
 	/* upgrade data layout before starting service if necessary*/
 	bool upgrade;
+	struct sd_stat stat;
 };
 
 struct siocb {
 	uint32_t epoch;
 	void *buf;
 	uint32_t length;
-	uint64_t offset;
+	uint32_t offset;
+	uint8_t ec_index;
+	uint8_t copy_policy;
 };
 
+/* This structure is used to pass parameters to vdi_* functions. */
 struct vdi_iocb {
 	const char *name;
 	const char *tag;
@@ -167,11 +183,16 @@ struct vdi_iocb {
 	uint32_t base_vid;
 	uint32_t snapid;
 	bool create_snapshot;
-	int nr_copies;
+	uint8_t copy_policy;
+	uint8_t store_policy;
+	uint8_t nr_copies;
+	uint64_t time;
 };
 
+/* This structure is used to get information from sheepdog. */
 struct vdi_info {
 	uint32_t vid;
+	uint32_t snapid;
 	uint32_t free_bit;
 	uint64_t create_time;
 };
@@ -180,20 +201,21 @@ struct vdi_state {
 	uint32_t vid;
 	uint8_t nr_copies;
 	uint8_t snapshot;
-	uint16_t _pad;
+	uint8_t copy_policy;
+	uint8_t _pad;
 };
 
 struct store_driver {
-	struct list_head list;
+	struct list_node list;
 	const char *name;
 	int (*init)(void);
-	bool (*exist)(uint64_t oid);
+	bool (*exist)(uint64_t oid, uint8_t ec_index);
 	/* create_and_write must be an atomic operation*/
 	int (*create_and_write)(uint64_t oid, const struct siocb *);
 	int (*write)(uint64_t oid, const struct siocb *);
 	int (*read)(uint64_t oid, const struct siocb *);
 	int (*format)(void);
-	int (*remove_object)(uint64_t oid);
+	int (*remove_object)(uint64_t oid, uint8_t ec_index);
 	int (*get_hash)(uint64_t oid, uint32_t epoch, uint8_t *sha1);
 	/* Operations in recovery */
 	int (*link)(uint64_t oid, uint32_t tgt_epoch);
@@ -203,8 +225,11 @@ struct store_driver {
 	int (*cleanup)(void);
 };
 
+/* backend store */
+int peer_read_obj(struct request *req);
+
 int default_init(void);
-bool default_exist(uint64_t oid);
+bool default_exist(uint64_t oid, uint8_t ec_index);
 int default_create_and_write(uint64_t oid, const struct siocb *iocb);
 int default_write(uint64_t oid, const struct siocb *iocb);
 int default_read(uint64_t oid, const struct siocb *iocb);
@@ -212,15 +237,18 @@ int default_link(uint64_t oid, uint32_t tgt_epoch);
 int default_update_epoch(uint32_t epoch);
 int default_cleanup(void);
 int default_format(void);
-int default_remove_object(uint64_t oid);
+int default_remove_object(uint64_t oid, uint8_t ec_index);
 int default_get_hash(uint64_t oid, uint32_t epoch, uint8_t *sha1);
 int default_purge_obj(void);
-int for_each_object_in_wd(int (*func)(uint64_t, char *, uint32_t, void *), bool,
-			  void *);
-int for_each_object_in_stale(int (*func)(uint64_t oid, char *path,
-					 uint32_t epoch, void *arg),
+int for_each_object_in_wd(int (*func)(uint64_t, const char *, uint32_t,
+				      uint8_t, void *),
+			  bool, void *);
+int for_each_object_in_stale(int (*func)(uint64_t oid, const char *path,
+					 uint32_t epoch, uint8_t, void *arg),
 			     void *arg);
-int for_each_obj_path(int (*func)(char *path));
+int for_each_obj_path(int (*func)(const char *path));
+size_t get_store_objsize(uint64_t oid);
+int get_store_path(uint64_t oid, uint8_t ec_index, char *path);
 
 extern struct list_head store_drivers;
 #define add_store_driver(driver)				\
@@ -256,8 +284,9 @@ static inline bool is_aligned_to_pagesize(void *p)
 	return ((uintptr_t)p & (getpagesize() - 1)) == 0;
 }
 
-int create_listen_port(char *bindaddr, int port);
+int create_listen_port(const char *bindaddr, int port);
 int init_unix_domain_socket(const char *dir);
+void unregister_listening_fds(void);
 
 int init_store_driver(bool is_gateway);
 int init_global_pathnames(const char *d, char *);
@@ -268,15 +297,20 @@ int lock_base_dir(const char *d);
 int fill_vdi_state_list(void *data);
 bool oid_is_readonly(uint64_t oid);
 int get_vdi_copy_number(uint32_t vid);
+int get_vdi_copy_policy(uint32_t vid);
 int get_obj_copy_number(uint64_t oid, int nr_zones);
-int get_max_copy_number(void);
 int get_req_copy_number(struct request *req);
-int add_vdi_state(uint32_t vid, int nr_copies, bool snapshot);
+int add_vdi_state(uint32_t vid, int nr_copies, bool snapshot, uint8_t);
 int vdi_exist(uint32_t vid);
-int vdi_create(struct vdi_iocb *iocb, uint32_t *new_vid);
-int vdi_delete(struct vdi_iocb *iocb, struct request *req);
-int vdi_lookup(struct vdi_iocb *iocb, struct vdi_info *info);
+int vdi_create(const struct vdi_iocb *iocb, uint32_t *new_vid);
+int vdi_snapshot(const struct vdi_iocb *iocb, uint32_t *new_vid);
+int vdi_delete(const struct vdi_iocb *iocb, struct request *req);
+int vdi_lookup(const struct vdi_iocb *iocb, struct vdi_info *info);
 void clean_vdi_state(void);
+int sd_delete_vdi(const char *name);
+int sd_lookup_vdi(const char *name, uint32_t *vid);
+
+extern int ec_max_data_strip;
 
 int read_vdis(char *data, int len, unsigned int *rsp_len);
 
@@ -290,8 +324,7 @@ int local_get_node_list(const struct sd_req *req, struct sd_rsp *rsp,
 struct vnode_info *grab_vnode_info(struct vnode_info *vnode_info);
 struct vnode_info *get_vnode_info(void);
 void put_vnode_info(struct vnode_info *vinfo);
-struct vnode_info *alloc_vnode_info(const struct sd_node *nodes,
-				    size_t nr_nodes);
+struct vnode_info *alloc_vnode_info(const struct rb_root *);
 struct vnode_info *get_vnode_info_epoch(uint32_t epoch,
 					struct vnode_info *cur_vinfo);
 void wait_get_vdis_done(void);
@@ -314,10 +347,11 @@ int inc_and_log_epoch(void);
 
 extern char *config_path;
 int set_cluster_config(const struct cluster_info *cinfo);
-int get_cluster_config(struct cluster_info *cinfo);
 int set_node_space(uint64_t space);
 int get_node_space(uint64_t *space);
 bool is_cluster_formatted(void);
+bool was_cluster_shutdowned(void);
+int set_cluster_shutdown(bool);
 
 int store_file_write(void *buffer, size_t len);
 void *store_file_read(void);
@@ -340,14 +374,19 @@ void get_recovery_state(struct recovery_state *state);
 
 int read_backend_object(uint64_t oid, char *data, unsigned int datalen,
 		       uint64_t offset);
-int write_object(uint64_t oid, char *data, unsigned int datalen,
-		 uint64_t offset, bool create);
-int read_object(uint64_t oid, char *data, unsigned int datalen,
-		uint64_t offset);
-int remove_object(uint64_t oid);
+int sd_write_object(uint64_t oid, char *data, unsigned int datalen,
+		    uint64_t offset, bool create);
+int sd_read_object(uint64_t oid, char *data, unsigned int datalen,
+		   uint64_t offset);
+int sd_remove_object(uint64_t oid);
+int sd_discard_object(uint64_t oid);
 
+struct request_iocb *local_req_init(void);
 int exec_local_req(struct sd_req *rq, void *data);
-void local_req_init(void);
+int exec_local_req_async(struct sd_req *rq, void *, struct request_iocb *);
+int local_req_wait(struct request_iocb *iocb);
+
+void local_request_init(void);
 
 int prealloc(int fd, uint32_t size);
 
@@ -356,6 +395,17 @@ void objlist_cache_remove(uint64_t oid);
 
 void put_request(struct request *req);
 
+int sheep_bnode_writer(uint64_t oid, void *mem, unsigned int len,
+		       uint64_t offset, uint32_t flags, int copies,
+		       int copy_policy, bool create, bool direct);
+int sheep_bnode_reader(uint64_t oid, void **mem, unsigned int len,
+		       uint64_t offset);
+
+#define INODE_GET_VID(inode, idx) (sd_inode_get_vid(sheep_bnode_reader, \
+					inode, idx))
+#define INODE_SET_VID(inode, idx, vdi_id) (sd_inode_set_vid(sheep_bnode_writer,\
+					sheep_bnode_reader, inode, idx, vdi_id))
+
 /* Operations */
 
 const struct sd_op_template *get_sd_op(uint8_t opcode);
@@ -365,17 +415,19 @@ bool is_local_op(const struct sd_op_template *op);
 bool is_peer_op(const struct sd_op_template *op);
 bool is_gateway_op(const struct sd_op_template *op);
 bool is_force_op(const struct sd_op_template *op);
+bool is_logging_op(const struct sd_op_template *op);
 bool has_process_work(const struct sd_op_template *op);
 bool has_process_main(const struct sd_op_template *op);
 void do_process_work(struct work *work);
 int do_process_main(const struct sd_op_template *op, const struct sd_req *req,
 		    struct sd_rsp *rsp, void *data);
-int sheep_do_op_work(const struct sd_op_template *op, struct request *req);
 int gateway_to_peer_opcode(int opcode);
 
+extern uint32_t last_gathered_epoch;
+
 static inline bool vnode_is_local(const struct sd_vnode *v)
 {
-	return node_id_cmp(&v->nid, &sys->this_node.nid) == 0;
+	return node_id_cmp(&v->node->nid, &sys->this_node.nid) == 0;
 }
 
 static inline bool node_is_local(const struct sd_node *n)
@@ -388,12 +440,7 @@ int gateway_read_obj(struct request *req);
 int gateway_write_obj(struct request *req);
 int gateway_create_and_write_obj(struct request *req);
 int gateway_remove_obj(struct request *req);
-
-/* backend store */
-int peer_read_obj(struct request *req);
-int peer_write_obj(struct request *req);
-int peer_create_and_write_obj(struct request *req);
-int peer_remove_obj(struct request *req);
+bool is_erasure_oid(uint64_t oid);
 
 /* object_cache */
 
@@ -401,7 +448,6 @@ void object_cache_format(void);
 bool bypass_object_cache(const struct request *req);
 bool object_is_cached(uint64_t oid);
 
-void object_cache_try_to_reclaim(int);
 int object_cache_handle_request(struct request *req);
 int object_cache_write(uint64_t oid, char *data, unsigned int datalen,
 		       uint64_t offset, bool create);
@@ -431,22 +477,42 @@ journal_write_store(uint64_t oid, const char *buf, size_t size, off_t, bool);
 int journal_remove_object(uint64_t oid);
 
 /* md.c */
-bool md_add_disk(char *path);
+bool md_add_disk(const char *path, bool);
 uint64_t md_init_space(void);
-char *md_get_object_path(uint64_t oid);
-int md_handle_eio(char *);
-bool md_exist(uint64_t oid);
-int md_get_stale_path(uint64_t oid, uint32_t epoch, char *path);
+const char *md_get_object_dir(uint64_t oid);
+int md_handle_eio(const char *);
+bool md_exist(uint64_t oid, uint8_t ec_index);
+int md_get_stale_path(uint64_t oid, uint32_t epoch, uint8_t ec_index, char *);
 uint32_t md_get_info(struct sd_md_info *info);
 int md_plug_disks(char *disks);
 int md_unplug_disks(char *disks);
 uint64_t md_get_size(uint64_t *used);
 
+static inline bool is_stale_path(const char *path)
+{
+	return !!strstr(path, ".stale");
+}
+
+static inline bool is_stale_dentry(const char *dentry)
+{
+	return !!strstr(dentry, ".");
+}
+
+static inline bool is_tmp_dentry(const char *dentry)
+{
+	return !!strstr(dentry, ".tmp");
+}
+
+static inline bool is_ec_dentry(const char *dentry)
+{
+	return !!strstr(dentry, "_");
+}
+
 /* http.c */
 #ifdef HAVE_HTTP
-int http_init(const char *address);
+int http_init(const char *options);
 #else
-static inline int http_init(const char *address)
+static inline int http_init(const char *options)
 {
 	sd_notice("http service is not complied");
 	return 0;
